/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of he MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 * 
 * This Source Code is part of the More-CPP CORE library.
 * https://www.more-cpp.com
 */
#include <stdio.h>

#include "../../common/utils/utils.h"
#include "../../common/debug/log.h"
#include "../../common/resources/fs/object.h"
#include "../../common/data/buffer.h"
#include "../../common/resources/storage/container.h"
#include "../../common/resources/storage/base.h"
#include "../../common/utils/types/endian.h"
#include "../../common/utils/types/type.h"

// display help
void help(void)
{
	printf("Usage:" EOL);
	printf("\tspc dump \"filename\"" EOL);
	printf(EOL);
}

// recursive print of storage objects
void rprint(const core_resources_filesystem_object_base_t& obj, bool recursive, int depth = 0)
{
	// print current
	for (int i = 0; i < depth; i++)
		printf("   |");

	printf("   + %s %s::%s" EOL, obj.get_type_name().c_str(), obj.get_owner_name().c_str(), obj.get_var_name().c_str());

	// skip rest if non-recursive
	if (!recursive)
		return;

	// print vars
	auto vars = obj.get_vars();

	for (auto& p : vars)
	{
		if (p == nullptr)
			continue;

		for (int i = 0; i < depth + 1; i++)
			printf("   |");

		printf("   + %s %s::%s" EOL, p->type.c_str(), p->owner.c_str(), p->name.c_str());
	}

	// print sub-objects
	auto children = obj.get_children();

	for (auto& v : children)
		rprint(*v, recursive, depth + 1);
}

// dump containers of file
int dump(int argc, char** argv)
{
	if (argc < 1)
		return 1;

	const char* filename = argv[0];

	// load file
	printf("Reading file %s..." EOL, filename);

	core_resources_storage_container_t container;
	auto fs = container.unpack_ex(core::load_buffer_from_file(filename));

	printf(EOL);

	// retrieve list of objects
	auto list = fs->get_list();

	auto object = fs->create_object();

	if (object == nullptr)
	{
		printf("cannot create reader object!" EOL);

		return 1;
	}

	// get number of containers
	if (list.empty())
		printf("file is empty" EOL);
	else
	{
		printf("file has %zu objects(s):" EOL, list.size());

		for (auto& buffer : list)
		{
			// unpack to object
			object->unpack(buffer);

			// print
			rprint(*object, true);
		}
	}

	return 0;
}

// main program
int main(int argc, char** argv)
{
	try
	{
		if (argc < 2)
		{
			help();

			return 1;
		}

		if (_strcmpi(argv[1], "dump") == 0)
			return dump(argc - 2, argv + 2);

		printf("Unknown command \"%s\"" EOL, argv[1]);

		help();
	}
	catch (core_exception_base_t& e)
	{
		printf("error: %s" EOL, e.to_string().c_str());

		_critical("%s", e.to_string().c_str());
	}
	catch (std::exception& e)
	{
		printf("error: %s" EOL, e.what());

		_critical("%s", e.what());
	}
	catch (...)
	{
		printf("unknown error!" EOL);

		_critical("unknown error!");
	}

	// signal error
	return 1;
}