/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of he MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 * 
 * This Source Code is part of the More-CPP CORE library.
 * https://www.more-cpp.com
 */
#include "../../common/utils/utils.h"
#include "../../common/debug/log.h"
#include "../../common/exception.h"
#include "../../common/arch/instancers/instancer.h"
#include "../../common/arch/library/library.h"
#include "../../common/utils/test/test.h"

#include <list>
#include <regex>

#include <stdio.h>
#include <stdlib.h>

// main program
int main(int argc, char** argv)
{
	// parameters
	bool verbose = false;

	// exclusion list
	std::list<std::regex> exclude_list;

	auto _exclude = [&](const std::string& name)
	{
		for (auto& v : exclude_list)
			if (std::regex_match(name, v))
				return true;

		return false;
	};

	// filter list
	std::list<std::regex> filter_list;

	auto _filter = [&](const std::string& name)
	{
		if (filter_list.empty())
			return false;

		for (auto& v : filter_list)
			if (std::regex_match(name, v))
				return false;

		return true;
	};

	// browse arguments
	for (int i = 1; i < argc; i++)
	{
		if ((_strcmpi(argv[i], "-e") == 0 || _strcmpi(argv[i], "--exclude") == 0) && (i + 1) < argc)
			exclude_list.push_back(std::regex(argv[++i]));
		else if ((_strcmpi(argv[i], "-f") == 0 || _strcmpi(argv[i], "--filter") == 0) && (i + 1) < argc)
			filter_list.push_back(std::regex(argv[++i]));
		else if (_strcmpi(argv[i], "-v") == 0 || _strcmpi(argv[i], "--verbose") == 0)
			verbose = true;
		else
		{
			printf("Unknown command %s!" EOL, argv[i]);

			return 1;
		}
	}

	try
	{
		// load all libraries
		core::load_libraries_standard();

		// use repeatable seed
		srand(0xfeed);

		// populate tests from all libraries
		std::list<std::string> list;

		auto lib_list = core::get_instance<core_arch_library_manager_t>()->list();

		for (auto& curr_lib : lib_list)
		{
			if (verbose)
				_verbose("adding tests from %s", curr_lib.c_str());

			try
			{
				// get remote singleton
				auto p = core::get_remote_singleton<core_arch_instancer_global_t<core_utils_test_base_t>>(curr_lib);

				if (p != nullptr)
				{
					// get list
					auto sub_list = p->list();

					if (verbose)
						_verbose("importing %zu tests", sub_list.size());

					// insert into main list
					for (auto& curr_test : sub_list)
						list.push_back(curr_lib + "/" + curr_test);
				}
			}
			catch (...) {}
		}

		// run all tests
		size_t num = 0;

		for (auto& v : list)
		{
			// skip if in list
			if (_exclude(v) || _filter(v))
			{
				if (verbose)
					_verbose("skipping test %s", v.c_str());

				continue;
			}

			// run test
			if (verbose)
				_verbose("running test %s", v.c_str());

			auto p = core::create_instance<core_utils_test_base_t>(v);

			if (__ISNULLPTR(p))
				continue;

			p->run();

			num++;
		}

		// return successfuly
		return 0;
	}
	catch (core_exception_base_t& e)
	{
		printf("error: %s" EOL, e.to_string().c_str());

		_critical("%s", e.to_string().c_str());
	}
	catch (std::exception& e)
	{
		printf("error: %s" EOL, e.what());

		_critical("%s", e.what());
	}
	catch (...)
	{
		printf("unknown error!" EOL);

		_critical("unknown error!");
	}

	// signal error
	return 1;
}
