/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of he MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 * 
 * This Source Code is part of the More-CPP CORE library.
 * https://www.more-cpp.com
 */
#pragma once

#include "../../utils/utils.h"
#include "../../debug/exdebug.h"
#include "../../debug/log.h"

#include "../singleton.h"

#include "exception.h"
#include "base.h"

#include <string>
#include <memory>
#include <unordered_map>

namespace core::arch::interface
{
	/* library interface manager singleton class */
	class manager : public core_arch_singleton_t<manager>
	{
		friend class singleton<manager>;

	public:

		// dtor
		~manager(void)
		{
			_debug("destroying library interface manager");

			this->m_interfaces.clear();
		}

		// check if interface exists
		template<class type> bool has(void) const
		{
			auto type_name = std::string(typeid(type).name());

			// check if existing
			return this->m_interfaces.find(type_name) != this->m_interfaces.end();
		}

		// retrieve current interface for type
		template<class type> std::shared_ptr<const type> get(void) const
		{
			auto type_name = std::string(typeid(type).name());

			// check if existing
			auto it = this->m_interfaces.find(type_name);

			if (it == this->m_interfaces.end())
				throw_exception(unknown_interface_exception, type_name);

			// get casted pointer
			auto p = std::static_pointer_cast<const type>(it->second);

			// throw exception on invalid interface
			if (__ISNULLPTR(p))
				throw_exception(null_interface_exception);

			return p;
		}

		// change current interface for type
		template<class type> void set(std::shared_ptr<const type> p)
		{
			// throw exception on invalid interface
			if (__ISNULLPTR(p))
				throw_exception(null_interface_exception);

			// get type name
			auto type_name = std::string(typeid(type).name());

			// check if already existing
			auto it = this->m_interfaces.find(type_name);

			if (it != this->m_interfaces.end())
			{
				it->second = p;

				return;
			}

			// insert into list
			_debug("registering interface \"%s\"", type_name.c_str());

			this->m_interfaces.emplace(std::make_pair(type_name, p));
		}

		// detach all interfaces
		void detach_all(void)
		{
			this->m_interfaces.clear();
		}

		// auto_register interface
		template<class type> class auto_register
		{
		public:
			auto_register(std::shared_ptr<const type> p)
			{
				core::get_instance<core::arch::interface::manager>()->set<type>(p);
			}
		};

	private:
		std::unordered_map<std::string, std::shared_ptr<const base>> m_interfaces;
	};
};

// shorthands
using core_arch_interface_manager_t = core::arch::interface::manager;