/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of he MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 * 
 * This Source Code is part of the More-CPP CORE library.
 * https://www.more-cpp.com
 */
#pragma once

#include "../singleton_manager.h"

#include "../../debug/exdebug.h"

#include "manager.h"

#include <filesystem>

#ifdef _WIN32
#include <Windows.h>
#endif

namespace core
{
	// retrieve master library manager
	static core_arch_library_manager_t* get_base_library_manager(void)
	{
		// get singleton manager of base process
		auto p = get_instance<core_arch_library_manager_t>()->get_addr(".", "get_singleton_manager");

		if (__ISNULLPTR(p))
			throw_exception(cannot_load_singleton_manager_exception);

		// resolve singleton manager
		auto manager = (*((core_arch_singleton_manager_t * (*)(void))p))();

		if (__ISNULLPTR(manager))
			throw_exception(cannot_load_singleton_manager_exception);

		// get library manager
		return static_cast<core_arch_library_manager_t*>(manager->get(typeid(core_arch_library_manager_t).name()));
	}

	// get remote singleton
	template<class type> type* get_remote_singleton(const std::string& library)
	{
		// get singleton manager func addr
		auto p = get_base_library_manager()->get_addr(library, "get_singleton_manager");

		if (__ISNULLPTR(p))
			throw_exception(cannot_load_singleton_manager_exception);

		// resolve singleton manager
		auto manager = (*((core_arch_singleton_manager_t* (*)(void))p))();

		if (__ISNULLPTR(manager))
			throw_exception(cannot_load_singleton_manager_exception);

		// get requested singleton
		return static_cast<type*>(manager->get(typeid(type).name()));
	}

	// check if remote singleton exists
	template<class type> bool has_remote_singleton(const std::string& library)
	{
		// get singleton manager func addr
		auto p = get_base_library_manager()->get_addr(library, "get_singleton_manager");

		if (__ISNULLPTR(p))
			return false;

		// resolve singleton manager
		auto manager = (*((core_arch_singleton_manager_t* (*)(void))p))();

		if (__ISNULLPTR(manager))
			throw_exception(cannot_load_singleton_manager_exception);

		// get requested singleton
		return manager->has(typeid(type).name());
	}

	// load all libraries from that folder
	static void load_libraries_from_path(const std::string& path)
	{
		_debug("loading all libraries from path %s", path.c_str());

		// load libraries
		for (auto const& curr_entry : std::filesystem::directory_iterator(path))
		{
			if (!curr_entry.is_regular_file())
				continue;

			auto filename = curr_entry.path().string();

#if _WIN32
			if (_strcmpi(split_file_parts(filename).ext.c_str(), "dll") != 0)
				continue;
#else
			if (_strcmpi(split_file_parts(filename).ext.c_str(), "so") != 0)
				continue;
#endif

			get_instance<core_arch_library_manager_t>()->load(filename);
		}
	}

	// standard load of libraries
	static void load_libraries_standard(void)
	{
		// load from current directory
		load_libraries_from_path("./");

		// load from application directory on Windows
#ifdef _WIN32
		char path[MAX_PATH];

		GetModuleFileNameA(NULL, path, MAX_PATH);

		load_libraries_from_path(split_file_path(path).directory);
#endif
	}
};