/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of he MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 * 
 * This Source Code is part of the More-CPP CORE library.
 * https://www.more-cpp.com
 */
#pragma once

#include "../../utils/utils.h"
#include "../../exception.h"
#include "../../debug/log.h"
#include "../../debug/exdebug.h"

#include "exception.h"
#include "base.h"

#include <string>

#ifdef _WIN32
#include <Windows.h>
#else
#include <dlfcn.h>
#endif

namespace core::arch::library
{
	/* shared object libraries */
	class so : public base
	{
	public:

		// acquire handle on ctor
		so(const std::string& filename)
		{
			_debug("loading shared object \"%s\"...", filename.c_str());

#ifdef _WIN32
			this->m_handle = LoadLibraryA(filename.c_str());

			if (this->m_handle == NULL)
				throw_exception(invalid_library_exception, filename);
#else
			this->m_handle = dlopen(filename.c_str(), RTLD_LAZY);

			if (this->m_handle == nullptr)
				throw_exception(invalid_library_exception, filename);
#endif
		}

		// free handle on dtor
		~so(void)
		{
			try
			{
				// call cleanup function if existing
				auto fn_cleanup = (void(*)(void))get_addr("cleanup");

				if (fn_cleanup != nullptr)
					(*fn_cleanup)();
			}
			catch (...) {}

#ifdef _WIN32
			if (this->m_handle != NULL)
				FreeLibrary(this->m_handle);

			this->m_handle = NULL;
#else
			if (this->m_handle != nullptr)
				dlclose(this->m_handle);

			this->m_handle = nullptr;
#endif
		}

		// get address
		virtual void* get_addr(const std::string& funcname) const override
		{
#ifdef _WIN32
			auto p = GetProcAddress(this->m_handle, funcname.c_str());
#else
			auto p = dlsym(this->m_handle, funcname.c_str());
#endif

			if (__ISNULLPTR(p))
				throw_exception(unknown_func_exception, funcname);

			return p;
		}

	private:

#ifdef _WIN32
		HMODULE m_handle;
#else
		void* m_handle;
#endif
	};
};

// shorthands
using core_arch_library_so_t = core::arch::library::so;