/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of he MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 * 
 * This Source Code is part of the More-CPP CORE library.
 * https://www.more-cpp.com
 */
#pragma once

#include "../../threads/lock.h"

#include <memory>
#include <list>
#include <mutex>
#include <algorithm>

namespace core::arch::notify
{
	class actor;

	/* observer class */
	class observer
	{
		// actor can touch our private parts...
		friend class actor;

	public:

		// actor has been updated
		virtual bool on_update(actor&) = 0;

		// dtor
		~observer(void);

		// number of observed actor
		size_t num_observed(void) const
		{
			AUTOLOCK_EX(this->m_lock, std::recursive_mutex);

			return this->m_list.size();
		}

	protected:

		// observe/unobserve given actor
		void observe(actor *p);
		void unobserve(actor* p);	

	private:
		mutable std::recursive_mutex m_lock;
		std::list<actor*> m_list;
	};

	/* actor class */
	class actor
	{
		// observer can touch our private parts...
		friend class observer;

	public:

		// notify all observers on destruction
		~actor(void)
		{
			// copy list
			std::list<observer*> listcpy;

			{
				std::lock_guard<std::recursive_mutex> lk(this->m_lock);
				listcpy = this->m_list;
			}

			for (auto& p : listcpy)
				if (p != nullptr)
					p->unobserve(this);
		}

		// signal all observers we have been changed
		size_t signal(void)
		{
			size_t num = 0;

			std::lock_guard<std::recursive_mutex> lk(this->m_lock);

			for (auto& p : this->m_list)
				if (p != nullptr && p->on_update(*this))
					num++;

			return num;
		}

		// number of observees
		size_t num_observees(void) const
		{
			AUTOLOCK_EX(this->m_lock, std::recursive_mutex);

			return this->m_list.size();
		}

	private:

		// register a new observer
		void register_observer(observer *p)
		{
			std::lock_guard<std::recursive_mutex> lk(this->m_lock);

			if (std::find(this->m_list.begin(), this->m_list.end(), p) != this->m_list.end())
				return;

			this->m_list.push_back(p);
		}

		// unregister an observer
		void unregister_observer(observer *p)
		{
			std::lock_guard<std::recursive_mutex> lk(this->m_lock);

			this->m_list.remove(p);
		}

	private:
		mutable std::recursive_mutex m_lock;
		std::list<observer*> m_list;
	};	
};

// shorthands
using core_arch_notify_actor_t = core::arch::notify::actor;
using core_arch_notify_observer_t = core::arch::notify::observer;