/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of he MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 * 
 * This Source Code is part of the More-CPP CORE library.
 * https://www.more-cpp.com
 */
#pragma once

#include "../../utils/utils.h"
#include "../../exception.h"
#include "../../threads/lock.h"

#include "../scope/access.h"
#include "../scope/base.h"

#include <functional>
#include <memory>
#include <list>
#include <unordered_map>
#include <mutex>

/* invalid_callback_exception exception class */
class invalid_callback_exception : public core_exception_base_t
{
public:
	virtual std::string to_string(void) const override
	{
		return "invalid callback!";
	}
};

namespace core::arch::notify
{
	class listenner;

	/* base class */
	class base : public scope::base
	{
		friend class listenner;

	public:
		using observant_t = std::function<void(void)>;
		using observant_list_t = std::list<observant_t>;
		using iterator_t = typename observant_list_t::iterator;

	private:

		// access for unregistration
		class token : public scope::access<base>
		{
		public:
			token(base* parent, const scope::ticket& handle, int event, iterator_t p) : scope::access<base>(parent, handle)
			{
				_debug("listenning to event %d", event);

				this->m_event = event;
				this->m_pointer = p;
			}

			void unregister_listenner(void)
			{
				_debug("unlistenning to event %d", this->m_event);

				try
				{
					get()->unregister_listenner(this->m_event, this->m_pointer);
				}
				catch (out_of_scope_exception&)
				{
					_warning("object accessed out of its scope");
				}
			}

		private:
			int m_event;
			iterator_t m_pointer;
		};

	protected:

		// called after every notifications
		virtual void on_notify(int) {}

		// notify event
		auto notify(int event)
		{
			size_t cnt = 0;

			{
				AUTOLOCK_EX(this->m_lock, std::recursive_mutex);

				// open list
				auto it = this->m_observers.find(event);

				// browse list and call functions
				if (it != this->m_observers.end())
				{
					for (auto& v : it->second)
						if (v)
						{
							v();

							cnt++;
						}
				}
			}

			// notify owner object
			on_notify(event);

			// return number of observer that were effectively notified
			return cnt;
		}

	private:

		// register observers
		token register_listenner(int event, observant_t obj)
		{
			// trigger error if callback is invalid
			if (!obj)
				throw_exception(invalid_callback_exception);

			AUTOLOCK_EX(this->m_lock, std::recursive_mutex);

			// check if entries already exists
			auto it = this->m_observers.find(event);

			// add to list if event is already listenned
			if (it != this->m_observers.end())
			{
				it->second.push_front(obj);

				return token(this, this->get_scope_ticket(), event, it->second.begin());
			}

			// create if no data yet
			this->m_observers.emplace(std::make_pair(event, observant_list_t({ })));

			return register_listenner(event, obj);
		}

		// unlisten event
		void unregister_listenner(int event, iterator_t p)
		{
			AUTOLOCK_EX(this->m_lock, std::recursive_mutex);

			// open list
			auto it = this->m_observers.find(event);

			if (it == this->m_observers.end())
				return;

			// remove from list
			it->second.erase(p);
		}

	private:

		mutable std::recursive_mutex m_lock;

		std::unordered_map<int, observant_list_t> m_observers;
	};

	/* listenner class */
	class listenner
	{
	public:
		~listenner(void)
		{
			NOTHROW(unlisten_all());
		}

		void unlisten_all(void)
		{
			for (auto& v : this->m_observees)
				v.unregister_listenner();
		}

		void listen(base& object, int event, base::observant_t callback)
		{
			this->m_observees.emplace_back(object.register_listenner(event, callback));
		}

	private:
		std::list<base::token> m_observees;
	};

	/* full implementation */
	class impl : public base, public listenner
	{
	public:
		void listen_self(int event, base::observant_t callback)
		{
			this->listen(*this, event, callback);
		}
	};
};

// shorthands
using core_arch_notify_base_t = core::arch::notify::base;
using core_arch_notify_listenner_t = core::arch::notify::listenner;
using core_arch_notify_impl_t = core::arch::notify::impl;

#define _SELF(class,func)	std::bind(&class::func, this)
