/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of he MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 * 
 * This Source Code is part of the More-CPP CORE library.
 * https://www.more-cpp.com
 */
#pragma once

#include "../../threads/lock.h"

#include "notify.h"

#include <memory>
#include <mutex>
#include <optional>

namespace core::arch::notify
{
	/* property class */
	template<typename type> class property : public base
	{
	public:
		enum
		{
			EVENT_CHANGED=0,
			EVENT_CLEARED,
		};

	public:

		// default ctor
		property(void) {}

		// typed ctor
		property(const type value)
		{
			this->set(value);
		}

		// copy ctor
		property(const property& obj)
		{
			this->operator=(obj);
		}

		// move ctor
		property(property&& obj)
		{
			this->operator=(std::move(obj));
		}

		// assignment operator
		const property<type>& operator=(const property& obj)
		{
			this->set(obj.get());

			return *this;
		}

		// assign by value
		const property<type>& operator=(const type value)
		{
			this->set(value);

			return *this;
		}

		// move operator
		const property<type>& operator=(property&& obj)
		{
			this->m_ptr = std::move(obj.m_ptr);

			// force event notification
			obj.clear();

			return *this;
		}

		// cast operator
		operator type(void) const
		{
			return this->get();
		}

		// clear
		void clear(void)
		{
			{
				AUTOLOCK(this->m_lock);

				this->m_ptr.reset();
			}

			// notify others
			this->notify(EVENT_CLEARED);
		}

		// set function
		void set(const type value)
		{
			{
				AUTOLOCK(this->m_lock);

				this->m_ptr = std::make_unique<type>(value);
			}

			// notify others
			this->notify(EVENT_CHANGED);
		}

		// try_get if data
		std::optional<type> try_get(void) const
		{
			AUTOLOCK(this->m_lock);

			if (this->m_ptr == nullptr)
				return {};

			return { *this->m_ptr };
		}

		// get function
		const type get(void) const
		{
			AUTOLOCK(this->m_lock);

			if (this->m_ptr == nullptr)
				throw_exception(null_pointer_exception);

			return *this->m_ptr;
		}

	private:
		mutable std::mutex m_lock;
		std::unique_ptr<type> m_ptr;
	};
};

// shorthands
template<typename type> using core_arch_notify_property_t = core::arch::notify::property<type>;