/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of he MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 * 
 * This Source Code is part of the More-CPP CORE library.
 * https://www.more-cpp.com
 */
#pragma once

#include "../debug/exdebug.h"
#include "../debug/log.h"

#include "../exception.h"

#include "singleton.h"

#include <string>
#include <unordered_map>

/* unknown_singleton_exception exception class */
class unknown_singleton_exception : public core_exception_base_t
{
public:
	unknown_singleton_exception(const std::string& type)
	{
		this->m_type = type;
	}

	virtual std::string to_string(void) const override
	{
		return "Unknown singleton \"" + this->m_type + "\" !";
	}

private:
	std::string m_type;
};

namespace core::arch
{
	/* singleton_manager singleton class */
	class singleton_manager : public core_arch_singleton_t<singleton_manager>
	{
	public:

		// dtor
		~singleton_manager(void)
		{
			_debug("destroying singleton manager");
		}

		// check if exists
		bool has(const std::string& type_name) const
		{
			return this->m_list.find(type_name) != this->m_list.end();
		}

		// add to list
		void add(const std::string& type_name, void* p)
		{
			// skip if already into list
			if (has(type_name))
				return;

			// add to list
			_debug("adding \"%s\" to singleton manager", type_name.c_str());

			this->m_list.emplace(std::make_pair(type_name, p));
		}

		// get from list
		void* get(const std::string& type_name) const
		{
			auto it = this->m_list.find(type_name);

			// throw exception not found
			if (it == this->m_list.end())
				throw_exception(unknown_singleton_exception, type_name);

			// return pointer
			return it->second;
		}

		// register object
		class auto_register
		{
		public:
			auto_register(const std::string& type_name, void* p)
			{
				core::get_instance<singleton_manager>()->add(type_name, p);
			}
		};

	private:
		std::unordered_map<std::string, void*> m_list;
	};
};

using core_arch_singleton_manager_t = core::arch::singleton_manager;

#define BIND_SINGLETON_EX(type, varname)		core::arch::singleton_manager::auto_register g_auto_register_##varname(typeid(type).name(), type::instance);
#define BIND_SINGLETON(type)					BIND_SINGLETON_EX(type, type)