/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of he MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 * 
 * This Source Code is part of the More-CPP CORE library.
 * https://www.more-cpp.com
 */
#pragma once

#include "../../exception.h"

#include "../buffer.h"

#include <string>
#include <array>
#include <functional>

#include <stdint.h>

namespace core::data::cryptography
{
	/* aes cipher class implemented from FIPS-197 standard */
	class aes
	{
	public:
		static constexpr size_t Nb = 4;

		using byte = uint8_t;
		using overflow_byte = uint16_t;
		using word = uint32_t;

		using state = std::array<byte, 4 * Nb>;

		/* unsupported_key_size_exception exception class */
		class unsupported_key_size_exception : public core_exception_base_t
		{
		public:
			virtual std::string to_string(void) const override
			{
				return "aes key size not supported!";
			}
		};

		/* invalid_message_size_exception exception class */
		class invalid_message_size_exception : public core_exception_base_t
		{
		public:
			virtual std::string to_string(void) const override
			{
				return "aes messages must be multiples of 128 bits!";
			}
		};

	private:

		// create array from bytes
		template<size_t N> static auto bytes2array(const unsigned char* key_data)
		{
			std::array<byte, N> ret;

			for (size_t i = 0; i < N; i++)
				ret[i] = key_data[i];

			return ret;
		}

		// return index given row and column
		static size_t sub2ind(size_t r, size_t c)
		{
			return r + 4 * c;
		};

		// create word from bytes
		static word make_word(byte a, byte b, byte c, byte d)
		{
			return (word)d | ((word)c << 8) | ((word)b << 16) | ((word)a << 24);
		}

		// split word into bytes
		static std::tuple<byte, byte, byte, byte> split_word(word w)
		{
			byte d = (byte)(w & 0xFF);
			byte c = (byte)((w >> 8) & 0xFF);
			byte b = (byte)((w >> 16) & 0xFF);
			byte a = (byte)((w >> 24) & 0xFF);

			return std::tuple(a, b, c, d);
		}

		// apply S-box to byte
		static byte sub_byte(byte b)
		{
			static const byte sbox[256] =
			{
				0x63, 0x7c, 0x77, 0x7b, 0xf2, 0x6b, 0x6f, 0xc5, 0x30, 0x01, 0x67, 0x2b, 0xfe, 0xd7, 0xab, 0x76,
				0xca, 0x82, 0xc9, 0x7d, 0xfa, 0x59, 0x47, 0xf0, 0xad, 0xd4, 0xa2, 0xaf, 0x9c, 0xa4, 0x72, 0xc0,
				0xb7, 0xfd, 0x93, 0x26, 0x36, 0x3f, 0xf7, 0xcc, 0x34, 0xa5, 0xe5, 0xf1, 0x71, 0xd8, 0x31, 0x15,
				0x04, 0xc7, 0x23, 0xc3, 0x18, 0x96, 0x05, 0x9a, 0x07, 0x12, 0x80, 0xe2, 0xeb, 0x27, 0xb2, 0x75,
				0x09, 0x83, 0x2c, 0x1a, 0x1b, 0x6e, 0x5a, 0xa0, 0x52, 0x3b, 0xd6, 0xb3, 0x29, 0xe3, 0x2f, 0x84,
				0x53, 0xd1, 0x00, 0xed, 0x20, 0xfc, 0xb1, 0x5b, 0x6a, 0xcb, 0xbe, 0x39, 0x4a, 0x4c, 0x58, 0xcf,
				0xd0, 0xef, 0xaa, 0xfb, 0x43, 0x4d, 0x33, 0x85, 0x45, 0xf9, 0x02, 0x7f, 0x50, 0x3c, 0x9f, 0xa8,
				0x51, 0xa3, 0x40, 0x8f, 0x92, 0x9d, 0x38, 0xf5, 0xbc, 0xb6, 0xda, 0x21, 0x10, 0xff, 0xf3, 0xd2,
				0xcd, 0x0c, 0x13, 0xec, 0x5f, 0x97, 0x44, 0x17, 0xc4, 0xa7, 0x7e, 0x3d, 0x64, 0x5d, 0x19, 0x73,
				0x60, 0x81, 0x4f, 0xdc, 0x22, 0x2a, 0x90, 0x88, 0x46, 0xee, 0xb8, 0x14, 0xde, 0x5e, 0x0b, 0xdb,
				0xe0, 0x32, 0x3a, 0x0a, 0x49, 0x06, 0x24, 0x5c, 0xc2, 0xd3, 0xac, 0x62, 0x91, 0x95, 0xe4, 0x79,
				0xe7, 0xc8, 0x37, 0x6d, 0x8d, 0xd5, 0x4e, 0xa9, 0x6c, 0x56, 0xf4, 0xea, 0x65, 0x7a, 0xae, 0x08,
				0xba, 0x78, 0x25, 0x2e, 0x1c, 0xa6, 0xb4, 0xc6, 0xe8, 0xdd, 0x74, 0x1f, 0x4b, 0xbd, 0x8b, 0x8a,
				0x70, 0x3e, 0xb5, 0x66, 0x48, 0x03, 0xf6, 0x0e, 0x61, 0x35, 0x57, 0xb9, 0x86, 0xc1, 0x1d, 0x9e,
				0xe1, 0xf8, 0x98, 0x11, 0x69, 0xd9, 0x8e, 0x94, 0x9b, 0x1e, 0x87, 0xe9, 0xce, 0x55, 0x28, 0xdf,
				0x8c, 0xa1, 0x89, 0x0d, 0xbf, 0xe6, 0x42, 0x68, 0x41, 0x99, 0x2d, 0x0f, 0xb0, 0x54, 0xbb, 0x16
			};

			return sbox[b];
		}

		// apply inverse S-box to byte
		static byte inv_sub_byte(byte b)
		{
			static const byte sbox[256] =
			{
				0x52, 0x09, 0x6a, 0xd5, 0x30, 0x36, 0xa5, 0x38, 0xbf, 0x40, 0xa3, 0x9e, 0x81, 0xf3, 0xd7, 0xfb,
				0x7c, 0xe3, 0x39, 0x82, 0x9b, 0x2f, 0xff, 0x87, 0x34, 0x8e, 0x43, 0x44, 0xc4, 0xde, 0xe9, 0xcb,
				0x54, 0x7b, 0x94, 0x32, 0xa6, 0xc2, 0x23, 0x3d, 0xee, 0x4c, 0x95, 0x0b, 0x42, 0xfa, 0xc3, 0x4e,
				0x08, 0x2e, 0xa1, 0x66, 0x28, 0xd9, 0x24, 0xb2, 0x76, 0x5b, 0xa2, 0x49, 0x6d, 0x8b, 0xd1, 0x25,
				0x72, 0xf8, 0xf6, 0x64, 0x86, 0x68, 0x98, 0x16, 0xd4, 0xa4, 0x5c, 0xcc, 0x5d, 0x65, 0xb6, 0x92,
				0x6c, 0x70, 0x48, 0x50, 0xfd, 0xed, 0xb9, 0xda, 0x5e, 0x15, 0x46, 0x57, 0xa7, 0x8d, 0x9d, 0x84,
				0x90, 0xd8, 0xab, 0x00, 0x8c, 0xbc, 0xd3, 0x0a, 0xf7, 0xe4, 0x58, 0x05, 0xb8, 0xb3, 0x45, 0x06,
				0xd0, 0x2c, 0x1e, 0x8f, 0xca, 0x3f, 0x0f, 0x02, 0xc1, 0xaf, 0xbd, 0x03, 0x01, 0x13, 0x8a, 0x6b,
				0x3a, 0x91, 0x11, 0x41, 0x4f, 0x67, 0xdc, 0xea, 0x97, 0xf2, 0xcf, 0xce, 0xf0, 0xb4, 0xe6, 0x73,
				0x96, 0xac, 0x74, 0x22, 0xe7, 0xad, 0x35, 0x85, 0xe2, 0xf9, 0x37, 0xe8, 0x1c, 0x75, 0xdf, 0x6e,
				0x47, 0xf1, 0x1a, 0x71, 0x1d, 0x29, 0xc5, 0x89, 0x6f, 0xb7, 0x62, 0x0e, 0xaa, 0x18, 0xbe, 0x1b,
				0xfc, 0x56, 0x3e, 0x4b, 0xc6, 0xd2, 0x79, 0x20, 0x9a, 0xdb, 0xc0, 0xfe, 0x78, 0xcd, 0x5a, 0xf4,
				0x1f, 0xdd, 0xa8, 0x33, 0x88, 0x07, 0xc7, 0x31, 0xb1, 0x12, 0x10, 0x59, 0x27, 0x80, 0xec, 0x5f,
				0x60, 0x51, 0x7f, 0xa9, 0x19, 0xb5, 0x4a, 0x0d, 0x2d, 0xe5, 0x7a, 0x9f, 0x93, 0xc9, 0x9c, 0xef,
				0xa0, 0xe0, 0x3b, 0x4d, 0xae, 0x2a, 0xf5, 0xb0, 0xc8, 0xeb, 0xbb, 0x3c, 0x83, 0x53, 0x99, 0x61,
				0x17, 0x2b, 0x04, 0x7e, 0xba, 0x77, 0xd6, 0x26, 0xe1, 0x69, 0x14, 0x63, 0x55, 0x21, 0x0c, 0x7d
			};

			return sbox[b];
		}

		static byte xtime(byte val)
		{
			// no overflow
			if ((val & (1 << 7)) == 0)
				return val << 1;

			// overflow
			auto _val = (overflow_byte)val << 1;
			auto _xor = (overflow_byte)0x1b;
			auto _mask = (overflow_byte)0xff;

			return (byte)((_val ^ _xor) & _mask);
		}

		// multiply bytes
		static byte mult_bytes(byte a, byte b)
		{
			byte ret = 0;
			byte x = a;

			while (b > 0)
			{
				if ((b & 1) == 1)
					ret ^= x;

				x = xtime(x);

				b >>= 1;
			}

			return ret;
		}

		// apply s-box to bytes
		static state sub_bytes(const state& in)
		{
			state out;

			for (size_t i = 0; i < in.size(); i++)
				out[i] = sub_byte(in[i]);

			return out;
		}

		// apply s-box inverse to bytes
		static state inv_sub_bytes(const state& in)
		{
			state out;

			for (size_t i = 0; i < in.size(); i++)
				out[i] = inv_sub_byte(in[i]);

			return out;
		}

		// shift rows of state
		static state shift_rows(const state& in)
		{
			// return shift value
			auto _shift = [](size_t r)
			{
				return r;
			};

			state out;

			// perform shift
			for (size_t r = 0; r < 4; r++)
				for (size_t c = 0; c < Nb; c++)
					out[sub2ind(r, c)] = in[sub2ind(r, (c + _shift(r)) % Nb)];

			return out;
		}

		// inverse shift rows of state
		static state inv_shift_rows(const state& in)
		{
			// return shift value
			auto _shift = [](size_t r)
			{
				return r;
			};

			state out;

			// perform shift
			for (size_t r = 0; r < 4; r++)
				for (size_t c = 0; c < Nb; c++)
					out[sub2ind(r, (c + _shift(r)) % Nb)] = in[sub2ind(r, c)];

			return out;
		}

		// mix columns of state
		static state mix_columns(const state& in)
		{
			state out;

			for (size_t c = 0; c < Nb; c++)
			{
				out[sub2ind(0, c)] = mult_bytes(0x02, in[sub2ind(0, c)]) ^ mult_bytes(0x03, in[sub2ind(1, c)]) ^ in[sub2ind(2, c)] ^ in[sub2ind(3, c)];
				out[sub2ind(1, c)] = in[sub2ind(0, c)] ^ mult_bytes(0x02, in[sub2ind(1, c)]) ^ mult_bytes(0x03, in[sub2ind(2, c)]) ^ in[sub2ind(3, c)];
				out[sub2ind(2, c)] = in[sub2ind(0, c)] ^ in[sub2ind(1, c)] ^ mult_bytes(0x02, in[sub2ind(2, c)]) ^ mult_bytes(0x03, in[sub2ind(3, c)]);
				out[sub2ind(3, c)] = mult_bytes(0x03, in[sub2ind(0, c)]) ^ in[sub2ind(1, c)] ^ in[sub2ind(2, c)] ^ mult_bytes(0x02, in[sub2ind(3, c)]);
			}

			return out;
		}

		// inverse mix columns of state
		static state inv_mix_columns(const state& in)
		{
			state out;

			for (size_t c = 0; c < Nb; c++)
			{
				out[sub2ind(0, c)] = mult_bytes(0x0e, in[sub2ind(0, c)]) ^ mult_bytes(0x0b, in[sub2ind(1, c)]) ^ mult_bytes(0x0d, in[sub2ind(2, c)]) ^ mult_bytes(0x09, in[sub2ind(3, c)]);
				out[sub2ind(1, c)] = mult_bytes(0x09, in[sub2ind(0, c)]) ^ mult_bytes(0x0e, in[sub2ind(1, c)]) ^ mult_bytes(0x0b, in[sub2ind(2, c)]) ^ mult_bytes(0x0d, in[sub2ind(3, c)]);
				out[sub2ind(2, c)] = mult_bytes(0x0d, in[sub2ind(0, c)]) ^ mult_bytes(0x09, in[sub2ind(1, c)]) ^ mult_bytes(0x0e, in[sub2ind(2, c)]) ^ mult_bytes(0x0b, in[sub2ind(3, c)]);
				out[sub2ind(3, c)] = mult_bytes(0x0b, in[sub2ind(0, c)]) ^ mult_bytes(0x0d, in[sub2ind(1, c)]) ^ mult_bytes(0x09, in[sub2ind(2, c)]) ^ mult_bytes(0x0e, in[sub2ind(3, c)]);
			}

			return out;
		}

		// add a round of key to state
		static state add_round_key(const state& in, const state& key)
		{
			state out;

			// apply to all elements
			for (size_t i = 0; i < in.size(); i++)
				out[i] = in[i] ^ key[i];

			return out;
		}

		// apply S-box to word
		static word sub_word(word w)
		{
			// split word into bytes
			auto [a, b, c, d] = split_word(w);

			// return word of substituded bytes
			return make_word(sub_byte(a), sub_byte(b), sub_byte(c), sub_byte(d));
		}

		// rotate word
		static word rot_word(word w)
		{
			// split word into bytes
			auto [a, b, c, d] = split_word(w);

			// return word of scrambled bytes
			return make_word(b, c, d, a);
		}

		// key expansion scheme
		template<size_t Nr, size_t Nk> static auto key_expansion(const std::array<byte, 4 * Nk>& key)
		{
			std::array<word, Nb* (Nr + 1)> w;

			static const word rcon[] = { 0x01000000, 0x02000000, 0x04000000, 0x08000000, 0x10000000, 0x20000000, 0x40000000, 0x80000000, 0x1b000000, 0x36000000 };

			for (size_t i = 0; i < Nk; i++)
				w[i] = make_word(key[4 * i], key[4 * i + 1], key[4 * i + 2], key[4 * i + 3]);

			for (size_t i = Nk; i < Nb * (Nr + 1); i++)
			{
				auto temp = w[i - 1];

				if ((i % Nk) == 0)
					temp = sub_word(rot_word(temp)) ^ rcon[i / Nk - 1];
				else if ((i % Nk) == 4)
				{
					if constexpr (Nk > 6)
						temp = sub_word(temp);
				}

				w[i] = w[i - Nk] ^ temp;
			}

			return w;
		}

		// convert key to state
		template<size_t Nr, size_t Nk> static auto key2state(const std::array<word, Nb* (Nr + 1)>& w, size_t pos)
		{
			state out;

			for (size_t c = 0; c < Nb; c++)
			{
				// split key into bytes
				auto [b0, b1, b2, b3] = split_word(w[pos * Nb + c]);

				// fill in state
				out[sub2ind(0, c)] = b0;
				out[sub2ind(1, c)] = b1;
				out[sub2ind(2, c)] = b2;
				out[sub2ind(3, c)] = b3;
			}

			return out;
		}

		// encode block
		template<size_t Nr, size_t Nk> static auto encode(const state& in, const std::array<word, Nb* (Nr + 1)>& w)
		{
			// set input as current state
			auto curr = in;

			// add round of key
			curr = add_round_key(curr, key2state<Nr, Nk>(w, 0));

			// apply all rounds
			for (size_t round = 1; round < Nr; round++)
			{
				curr = sub_bytes(curr);
				curr = shift_rows(curr);
				curr = mix_columns(curr);
				curr = add_round_key(curr, key2state<Nr, Nk>(w, round));
			}

			// apply final round
			curr = sub_bytes(curr);
			curr = shift_rows(curr);
			curr = add_round_key(curr, key2state<Nr, Nk>(w, Nr));

			// return state as output
			return curr;
		}

		// decode block
		template<size_t Nr, size_t Nk> static auto decode(const state& in, const std::array<word, Nb* (Nr + 1)>& w)
		{
			// set input as current state
			auto curr = in;

			// add round of key
			curr = add_round_key(curr, key2state<Nr, Nk>(w, Nr));

			// apply all rounds
			for (size_t round = 1; round < Nr; round++)
			{
				curr = inv_shift_rows(curr);
				curr = inv_sub_bytes(curr);
				curr = add_round_key(curr, key2state<Nr, Nk>(w, Nr - round));
				curr = inv_mix_columns(curr);
			}

			// apply final round
			curr = inv_shift_rows(curr);
			curr = inv_sub_bytes(curr);
			curr = add_round_key(curr, key2state<Nr, Nk>(w, 0));

			// return state as output
			return curr;
		}

		// process message
		template<size_t Nr, size_t Nk> static core_data_buffer_t process(std::function<state(const state&, const std::array<word, Nb* (Nr + 1)>&)> func, const unsigned char* key_data, const core_data_buffer_t& message)
		{
			core_data_buffer_t ret;

			// expand key
			auto w = key_expansion<Nr, Nk>(bytes2array<4 * Nk>(key_data));

			// loop through message
			for (size_t i = 0; i < message.size(); i += 16)
			{
				auto curr = func(bytes2array<16>(message.ptr<unsigned char>(i)), w);

				for (size_t j = 0; j < curr.size(); j++)
					ret.append<unsigned char>(curr[j], false);
			}

			// return composed message
			return ret;
		}

		public:
			// encode message
			static core_data_buffer_t encode(const core_data_buffer_t& key, const core_data_buffer_t& message)
			{
				if ((message.size() * 8) % 128 != 0)
					throw_exception(invalid_message_size_exception);

				switch (key.size() * 8)
				{
				case 128:
					return process<10, 4>([](const state& in, const std::array<word, Nb * 11>& w)
						{
							return aes::encode<10, 4>(in, w);
						}, key.data(), message);

				case 192:
					return process<12, 6>([](const state& in, const std::array<word, Nb * 13>& w)
						{
							return aes::encode<12, 6>(in, w);
						}, key.data(), message);

				case 256:
					return process<14, 8>([](const state& in, const std::array<word, Nb * 15>& w)
						{
							return aes::encode<14, 8>(in, w);
						}, key.data(), message);

				default:
					throw_exception(unsupported_key_size_exception);
				}
			}

			// decode message
			static core_data_buffer_t decode(const core_data_buffer_t& key, const core_data_buffer_t& message)
			{
				if ((message.size() * 8) % 128 != 0)
					throw_exception(invalid_message_size_exception);

				switch (key.size() * 8)
				{
				case 128:
					return process<10, 4>([](const state& in, const std::array<word, Nb * 11>& w)
						{
							return aes::decode<10, 4>(in, w);
						}, key.data(), message);

				case 192:
					return process<12, 6>([](const state& in, const std::array<word, Nb * 13>& w)
						{
							return aes::decode<12, 6>(in, w);
						}, key.data(), message);

				case 256:
					return process<14, 8>([](const state& in, const std::array<word, Nb * 15>& w)
						{
							return aes::decode<14, 8>(in, w);
						}, key.data(), message);

				default:
					throw_exception(unsupported_key_size_exception);
				}
			}
	};
};

// shorthands
using core_data_cryptography_aes_t = core::data::cryptography::aes;