/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of he MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 * 
 * This Source Code is part of the More-CPP CORE library.
 * https://www.more-cpp.com
 */
#pragma once

#include "../buffer.h"

#include <string>
#include <cstdlib>

/* ill_formed_base64_encoding_exception exception class */
class ill_formed_base64_encoding_exception : public core_exception_base_t
{
public:
	virtual std::string to_string(void) const override
	{
		return "ill formed base64 encoding!";
	}
};

namespace core::data::encoder
{
	class base64
	{
	public:
		static core_data_buffer_t decode(const std::string& message)
		{
			const char* table = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

			core_data_buffer_t data;

			// find value
			auto _find = [&](const char val)
			{
				for (uint32_t i = 0; i < strlen(table); i++)
					if (table[i] == val)
						return i;

				throw_exception(ill_formed_base64_encoding_exception);
			};

			// size must be multiple of 4
			if (message.length() % 4 != 0)
				throw_exception(ill_formed_base64_encoding_exception);

			// read chunks of 4 characters
			for (size_t i = 0; i < message.length(); i += 4)
			{
				uint32_t temp = 0;

				for (size_t j = 0; j < 4; j++)
				{
					auto n = i + j;

					// check for end of string (ill formed)
					if (n >= message.length())
						throw_exception(ill_formed_base64_encoding_exception);

					// check for padding
					auto c = message.at(n);

					if (c == '=')
					{
						size_t num_pads = 0;

						for (size_t q = n; q < message.length(); q++)
						{
							if (message.at(q) != '=')
								throw_exception(ill_formed_base64_encoding_exception);

							num_pads++;

							temp <<= 6;
						}

						// should not have more than 2 pads
						if (num_pads > 2)
							throw_exception(ill_formed_base64_encoding_exception);

						// always add the two first bytes
						data.append<unsigned char>((temp >> 16) & 0xff);

						// only add last bytes if 1 pad
						if (num_pads == 1)
							data.append<unsigned char>((temp >> 8) & 0xff);

						// return data
						return data;
					}

					// find character
					uint32_t value = _find(c);

					// shift and add to temp
					temp <<= 6;
					temp += value;
				}

				// add to data
				data.append<unsigned char>((temp >> 16) & 0xff);
				data.append<unsigned char>((temp >> 8) & 0xff);
				data.append<unsigned char>(temp & 0xff);
			}

			// return data
			return data;
		}

		static std::string encode(const core_data_buffer_t& buffer)
		{
			std::string ret;

			const char* table = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

			// get pointer for data
			auto data = buffer.data();

			// loop by chunks of 3 bytes
			for (size_t i = 0; i < buffer.size(); i += 3)
			{
				// get 4 words of 6 bits (= 24 bits)
				for (size_t n = 0; n < 4; n++)
				{
					unsigned char curr = 0;

					if ((8 * i + n * 6) >= (8 * buffer.size()))
					{
						switch (8 * (buffer.size() - i) / 6)
						{
						case 1:
							return ret + "==";

						case 2:
							return ret + "=";

						default:
							return ret;
						}
					}

					for (size_t k = 0; k < 6; k++)
					{
						auto ofs = 8 * i + n * 6 + k;

						if (ofs < (8 * buffer.size()))
							curr |= ((data[ofs / 8] >> (7 - (ofs % 8))) & 1) << (5 - k);
					}

					// should not happend
					if (curr >= strlen(table))
					{
						ret += "?";

						continue;
					}

					// add current word
					char szTmp[2] = { table[curr], '\0' };

					ret += szTmp;
				}
			}

			// return string
			return ret;
		}
	};
};

// shorthands
using core_data_encoder_base64_t = core::data::encoder::base64;