/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of he MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 * 
 * This Source Code is part of the More-CPP CORE library.
 * https://www.more-cpp.com
 */
#pragma once

#include "utils/utils.h"

#include <string>
#include <functional>
#include <memory>

 // this action will no throw exception in case of failure
#define NOTHROW(action) { try{ action; } catch(...){} }

// generic exception class
namespace core::exception
{
	class base
	{
	public:
		base(void) {}
		virtual ~base(void) {}

		virtual std::string to_string(void) const = 0;
	};
};

// shorthands
using core_exception_base_t = core::exception::base;

namespace core
{
	// extended debug when triggering exceptions
	void ex_debug(const std::string& type, std::function<void(void)> callback);

	template<typename type> void ex_debug(std::function<void(void)> callback)
	{
		ex_debug(typeid(type).name(), callback);
	}
};

// send message to event monitor
void hook_exception(const core_exception_base_t& e, const std::string& type, const char* filename, unsigned int line);

#define throw_exception(exception_type, ...)       { exception_type e{__VA_ARGS__}; hook_exception(e, typeid(exception_type).name(), __FILE__, __LINE__); throw e; }

void breakpoint(void);

#define EX_BREAKPOINT(type)	core::ex_debug<type>(&breakpoint);

/* unknown_exception exception class */
class unknown_exception : public core_exception_base_t
{
public:
	virtual std::string to_string(void) const override
	{
		return "Unknown exception!";
	}
};

/* null_pointer_exception exception class */
class null_pointer_exception : public core_exception_base_t
{
public:
	virtual std::string to_string(void) const override
	{
		return "Pointer is null!";
	}
};

/* division_by_zero_exception exception class */
class division_by_zero_exception : public core_exception_base_t
{
public:
	virtual std::string to_string(void) const override
	{
		return "Division by zero!";
	}
};

/* function not implemented for tamplate exception class */
class template_func_not_impl_exception : public core_exception_base_t
{
public:
	template_func_not_impl_exception(const std::string& type)
	{
		this->m_type = type;
	}

	virtual std::string to_string(void) const override
	{
		return "Function is not implemented for template type \"" + this->m_type + "\" !";
	}

private:
	std::string m_type;
};

/* name_already_exist_exception exception class */
class name_already_exist_exception : public core_exception_base_t
{
public:
	name_already_exist_exception(const std::string& name)
	{
		this->m_name = name;
	}

	virtual std::string to_string(void) const override
	{
		return "Name \"" + this->m_name + "\" already exists!";
	}

private:
	std::string m_name;
};

/* name_not_found_exception exception class */
class name_not_found_exception : public core_exception_base_t
{
public:
	name_not_found_exception(const std::string& name)
	{
		this->m_name = name;
	}

	virtual std::string to_string(void) const override
	{
		return "\"" + this->m_name + "\" was not found!";
	}

private:
	std::string m_name;
};

/* memory_allocation_exception exception class */
class memory_allocation_exception : public core_exception_base_t
{
public:
	memory_allocation_exception(const size_t size)
	{
		this->m_size = size;
	}

	virtual std::string to_string(void) const override
	{
		char szTmp[64];

		snprintf(szTmp, sizeof(szTmp), "%zu", this->m_size);

		return "Cannot allocate " + std::string(szTmp) + " bytes !";
	}

private:
	size_t m_size;
};

/* cannot_open_file_exception exception class */
class cannot_open_file_exception : public core_exception_base_t
{
public:
	cannot_open_file_exception(const std::string& filename)
	{
		this->m_filename = filename;
	}

	virtual std::string to_string(void) const override
	{
		return std::string("Cannot open file \"") + this->m_filename + std::string("\" !");
	}

private:
	std::string m_filename;
};