/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of he MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 * 
 * This Source Code is part of the More-CPP CORE library.
 * https://www.more-cpp.com
 */
#pragma once

#include "../../debug/safe.h"
#include "../../data/buffer.h"

#include "../../exception.h"

#include "../storage/base.h"
#include "../fs/object.h"

#include "../exception.h"

#include <string>
#include <vector>
#include <array>

namespace core::resources::ext
{
	// storeable version of std::string
	class buffer : public core_data_buffer_t, public core_resources_storage_base_t
	{
	public:
		using core_data_buffer_t::buffer;
		using core_data_buffer_t::operator=;

		virtual std::string get_typename(void) const override
		{
			return "ext::buffer";
		}

		// push to storage object
		virtual void push(core::resources::filesystem::object_base& obj) const
		{
			auto bytesize = size();

			// write size
			obj.push_var(get_typename(), "size", "size_t", sizeof(bytesize), (void*)&bytesize);

			// write data
			obj.push_var(get_typename(), "data", "unsigned char", MAKE_UMULT(bytesize, sizeof(unsigned char)), (void*)data());

			// set typename
			obj.set_type_name(get_typename());
		}

		// pop from storage object
		virtual void pop(const core::resources::filesystem::object_base& obj)
		{
			// check typename
			if (!is_type_name(obj.get_type_name(), get_typename()))
				throw_exception(wrong_type_exception, get_typename(), obj.get_type_name());

			// free existing data
			clear();

			// read size
			size_t bytesize;

			try
			{

				obj.pop_var(get_typename(), "size", "size_t", sizeof(bytesize), (void*)&bytesize);
			}
			catch (...)
			{
				throw_exception(unknown_var_exception);
			}

			// allocate space and read data
			auto data = (unsigned char*)malloc(bytesize);

			if (data == nullptr)
				throw_exception(memory_allocation_exception, bytesize);

			try
			{
				obj.pop_var(get_typename(), "data", "unsigned char", MAKE_UMULT(bytesize, sizeof(unsigned char)), data);
			}
			catch (...)
			{
				// free temporary data
				if (data != nullptr)
					free(data);

				throw_exception(unknown_var_exception);
			}

			// assign result
			this->operator=(core_data_buffer_t(data, bytesize));
		}
	};
};

// shorthands
using core_resources_ext_buffer_t = core::resources::ext::buffer;