/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of he MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 * 
 * This Source Code is part of the More-CPP CORE library.
 * https://www.more-cpp.com
 */
#pragma once

#include "../../debug/safe.h"

#include "../../exception.h"

#include "../storage/base.h"
#include "../fs/object.h"

#include "../exception.h"

#include <string>

namespace core::resources::ext
{
	// storeable version of std::string
	class string : public std::string, public core_resources_storage_base_t
	{
	public:
		using std::string::string;
		using std::string::operator=;

		virtual std::string get_typename(void) const override
		{
			return "class storage_string";
		}

		// push to storage object
		virtual void push(core_resources_filesystem_object_base_t& container) const
		{
			size_t bytesize = MAKE_UADD(length(), (size_t)1);

			// write size
			container.push_var(get_typename(), "size", "size_t", sizeof(bytesize), (void*)&bytesize);

			// write data
			container.push_var(get_typename(), "data", "char", MAKE_UMULT(bytesize, sizeof(char)), (void*)c_str());

			// set typename
			container.set_type_name(get_typename());
		}

		// pop from storage object
		virtual void pop(const core_resources_filesystem_object_base_t& container)
		{
			// check typename
			if (!core::is_type_name(container.get_type_name(), get_typename()))
				throw_exception(wrong_type_exception, get_typename(), container.get_type_name());

			size_t bytesize = 0;

			// read size
			container.pop_var(get_typename(), "size", "size_t", sizeof(bytesize), (void*)&bytesize);

			// allocate space and read data
			char* string = (char*)malloc(bytesize);

			if (string == nullptr)
				throw_exception(memory_allocation_exception, bytesize);

			try
			{
				container.pop_var(get_typename(), "data", "char", MAKE_UMULT(bytesize, sizeof(char)), string);

				this->operator=(std::string(string));
			}
			catch (...)
			{
				// free temporary string
				if (string != nullptr)
					free(string);

				string = nullptr;

				throw_exception(unknown_var_exception);
			}

			// free temporary string
			if (string != nullptr)
				free(string);

			string = nullptr;
		}
	};
};

// shorthands
using core_resources_ext_string_t = core::resources::ext::string;