/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of he MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 * 
 * This Source Code is part of the More-CPP CORE library.
 * https://www.more-cpp.com
 */
#pragma once

#include "../../utils/utils.h"
#include "../../utils/types/type.h"
#include "../../arch/library/manager.h"

#include "../../exception.h"

#include "../fs/object.h"

#include "exception.h"

#include <vector>
#include <string>
#include <ios>
#include <fstream>

namespace core::resources::storage
{
	// structure to hold storage list
	struct list_s
	{
		const char* owner;
		const char* name;
		const char* type;

		size_t bytes_size;
		size_t bytes_offset;

		bool is_object;
		size_t object_base_offset;
	};

	// base class
	class base
	{
	public:
		virtual ~base(void) {}

		virtual void push(core::resources::filesystem::object_base& obj) const = 0;
		virtual void pop(const core::resources::filesystem::object_base& obj) = 0;

		virtual std::string get_typename(void) const = 0;

		virtual void custom_push(core::resources::filesystem::object_base&) const { }
		virtual void custom_pop(const core::resources::filesystem::object_base&) { }

		void generic_push(core::resources::filesystem::object_base& obj, const void* base_address, struct list_s* list, size_t num_elems, bool use_custom) const;
		void generic_pop(const core::resources::filesystem::object_base& obj, const void* base_address, struct list_s* list, size_t num_elems, bool use_custom);
	};
};

// shorthands
using core_resources_storage_base_t = core::resources::storage::base;

namespace core
{
	// SFINAE to check if storeable object
	template<typename type> struct is_base_object_t
	{
		static bool check(core_resources_storage_base_t*)
		{
			return true;
		}

		static bool check(...)
		{
			return false;
		}

		static bool test(void)
		{
			return check(static_cast<type*>(nullptr));
		}
	};

	template<typename type> static bool is_base_object(const type&)
	{
		return is_base_object_t<type> ::test();
	}

	// get base offset of multiple-inheritance classes
	template<class child, class parent> size_t get_base_ofs(void)
	{
		child tmp;

		parent* p = static_cast<parent*>(&tmp);

		return (size_t)((unsigned char*)p - (unsigned char*)&tmp);
	}

	// SFINAE for offset to storeable object
	template<typename type> struct object_base_offset_t
	{
		static size_t get_ofs(core_resources_storage_base_t*)
		{
			return get_base_ofs<type, core_resources_storage_base_t>();
		}

		static size_t get_ofs(...)
		{
			return 0;
		}

		static size_t get(void)
		{
			type tmp;

			return get_ofs(&tmp);
		}
	};

	template<typename type> size_t object_base_offset(const type&)
	{
		return object_base_offset_t<type> ::get();
	}

	// visual studio doesn't like typeid(((class*)0)->var)
	template<typename type> const char* get_var_typename(const type&)
	{
		return typeid(type).name();
	}
};

#define NOARG

// macro to implement storage mechanisms
#define IMPLEMENT_STORAGE										static struct core::resources::storage::list_s storage_data[];						\
																																	\
																virtual void push(core_resources_filesystem_object_base_t& obj) const override;	\
																virtual void pop(const core_resources_filesystem_object_base_t& obj) override;		\
																virtual std::string get_typename(void) const override;


#define BEGIN_STORAGE_EX(class, decorator)						decorator struct core::resources::storage::list_s class::storage_data[] = 

#define BEGIN_STORAGE(class)									BEGIN_STORAGE_EX(class, NOARG)


#define DECLARE_STORAGE(class, var)								{ #class, #var, core::get_var_typename(((class*)0)->var), sizeof(class::var), (size_t)&((class*)0)->var, core::is_base_object(((class*)0)->var), core::object_base_offset(((class*)0)->var) }

#define END_STORAGE_EX(class, parent, type_name, decorator)				;																			\
																																					\
																decorator void class::push(core_resources_filesystem_object_base_t& obj) const						\
																{																					\
																	if(strcmp(#class, #parent))														\
																		parent :: push(obj);														\
																																					\
																	size_t num_elems = sizeof(storage_data) / sizeof(struct core::resources::storage::list_s);		\
																																					\
																	generic_push(obj, this, storage_data, num_elems, _strcmpi(#class, #parent) == 0);\
																}																					\
																																					\
																decorator void class::pop(const core_resources_filesystem_object_base_t& obj)						\
																{																					\
																	if(strcmp(#class, #parent))														\
																		parent :: pop(obj);															\
																																					\
																	size_t num_elems = sizeof(storage_data) / sizeof(struct core::resources::storage::list_s);		\
																																					\
																	generic_pop(obj, this, storage_data, num_elems, _strcmpi(#class, #parent) == 0);\
																}																					\
																																					\
																decorator std::string class::get_typename(void) const								\
																{																					\
																	return core::decorate_typename(#class);												\
																}																					\
																																					\
																REGISTER_TYPENAME_EX(class, type_name, storage_type_##class);

#define END_STORAGE(class, parent)								END_STORAGE_EX(class, parent, #class, NOARG)

#define DECLARE_EMPTY_STORAGE_EX(class, parent, type_name, decorator)		decorator void class::push(core_resources_filesystem_object_base_t& obj) const			\
																{																					\
																	if(strcmp(#class, #parent))														\
																		parent :: push(obj);														\
																																					\
																	generic_push(obj, this, nullptr, 0, _strcmpi(#class, #parent) == 0);			\
																}																					\
																																					\
																decorator void class::pop(const core_resources_filesystem_object_base_t& obj)						\
																{																					\
																	if(strcmp(#class, #parent))														\
																		parent :: pop(obj);															\
																																					\
																	generic_pop(obj, this, nullptr, 0, _strcmpi(#class, #parent) == 0);				\
																}																					\
																																					\
																decorator std::string class::get_typename(void) const								\
																{																					\
																	return core::decorate_typename(#class);												\
																}																					\
																																					\
																REGISTER_TYPENAME_EX(class, type_name, storage_type_##class);


#define DECLARE_EMPTY_STORAGE(class, parent)					DECLARE_EMPTY_STORAGE_EX(class, parent, #class, NOARG)
