/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of he MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 * 
 * This Source Code is part of the More-CPP CORE library.
 * https://www.more-cpp.com
 */
#pragma once

#include "../arch/singleton.h"
#include "../data/encoder/base64.h"
#include "../data/hash/checksum.h"
#include "../data/hash/sha2.h"
#include "../data/buffer.h"

#include <atomic>

namespace core::resources::uid
{
	// create machine-based identification
	core_data_buffer_t ident(void);

	/* manager singleton class */
	class manager : public core_arch_singleton_t<manager>
	{
		friend class singleton<manager>;

	public:

		// dtor
		~manager(void)
		{
			_debug("destroying resources uid manager");
		}

		// return uid
		auto uid(void)
		{
			// copy ident
			core_data_buffer_t curr = this->m_ident;

			// add current time for randomness
			{
				auto currtime = std::chrono::time_point_cast<std::chrono::nanoseconds>(std::chrono::high_resolution_clock::now());

				curr.append(currtime.time_since_epoch().count());
			}

			// add thread-safe increment in case clock resolution is not high-enough
			{
				auto currinc = this->m_inc++;

				curr.append(currinc);
			}

			// return in base64
			return core_data_encoder_base64_t::encode(core::sha256(curr));
		}

		// return decorated uid
		auto create(void)
		{
			// format to string
			auto base64 = uid();

			char tmp[128];

			snprintf(tmp, sizeof(tmp), "%%{{%.8X$%s$%.8X}}%%", 'RND0', base64.c_str(), core::checksum32((const unsigned char*)base64.data(), base64.length()));

			return std::string(tmp);
		}

	private:
		manager(void)
		{
			this->m_inc = 0;

			this->m_ident = ident();
		}

		core_data_buffer_t m_ident;
		std::atomic<uint64_t> m_inc;
	};

	// create UID
	static auto create(void)
	{
		return core::get_instance<manager>()->create();
	}
};

// shorthands
using core_resources_uid_manager_t = core::resources::uid::manager;