/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of he MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 * 
 * This Source Code is part of the More-CPP CORE library.
 * https://www.more-cpp.com
 */
#pragma once

#include "../../utils/utils.h"
#include "../../debug/log.h"
#include "../../debug/exdebug.h"
#include "../../arch/singleton.h"

#include "../lock.h"
#include "../base.h"

#include "../pool/manager.h"

#include "exception.h"
#include "base.h"

namespace core::threads::jobs
{
	/* Jobs Manager class */
	class manager : public core_arch_singleton_t<manager>, public core_threads_base_t
	{
	public:
		friend class singleton<manager>;

		// clean exit
		virtual ~manager(void)
		{
			_debug("destroying jobs manager");

			try
			{
				stop();
			}
			catch (...) {}

			{
				AUTOLOCK(this->m_lock);

				this->m_jobs.clear();
			}
		}

		// check if empty
		bool empty(void)
		{
			AUTOLOCK(this->m_lock);

			return this->m_jobs.empty();
		}

		// add job
		void queue(std::shared_ptr<core_threads_jobs_base_t> p)
		{
			_debug("queuing job %p", p);

			// skip null jobs
			if (__ISNULLPTR(p))
				throw_exception(null_job_exception);

			// initialize job first if not already done
			if (p->get_status() == core_threads_jobs_base_t::status::init_required)
				p->init();

			// skip jobs that are not pending
			if (p->get_status() != core_threads_jobs_base_t::status::waiting)
				throw_exception(job_queue_exception);

			// push jobs to queue
			{

				AUTOLOCK(this->m_lock);

				p->push_job(this->m_jobs);
			}

			// sync mode
			if (!is_running())
			{
				// performs jobs as long as list is not empty
				while (!empty())
					run();
			}
		}

	private:

		// ctor
		manager(void) {}

		// inner queue
		void disptach(std::shared_ptr<core_threads_jobs_base_t> p);

		// run
		virtual void run(void)
		{
			// pick up a job
			auto p = pop();

			// wait 10 ms if stack is empty
			if (p == nullptr)
			{
				sleep(10e-3);

				return;
			}

			// assign to thread pool manager
			disptach(p);
		}

		// pick a job
		std::shared_ptr<core_threads_jobs_base_t> pop(void)
		{
			AUTOLOCK(this->m_lock);

			// browse list
			for (auto p : this->m_jobs)
			{
				// skip null
				if (__ISNULLPTR(p))
					continue;

				// skip if not ready to accept
				if (p->accept() != core_threads_jobs_base_t::ready::accept)
					continue;

				// adopt if waiting and remove from list
				if (p->get_status() == core_threads_jobs_base_t::status::waiting)
					p->adopt();

				this->m_jobs.remove(p);

				// return job
				return p;
			}

			// no jobs to process
			return nullptr;
		}

		// mutex
		lock_t m_lock;

		// list of jobs
		std::list<std::shared_ptr<core_threads_jobs_base_t>> m_jobs;
	};
};

// shorthands
using core_threads_jobs_manager_t = core::threads::jobs::manager;

namespace core
{
	// wait for
	static bool wait_for(std::list<std::shared_ptr<core_threads_jobs_base_t>> list, const double timeout = 0)
	{
		_debug("waiting for %zu jobs", list.size());

		// temporarily increase number of allowed threads
		auto obj = core::get_instance<core_threads_pool_manager_t>()->allow_more();

		// queue jobs
		for (auto& p : list)
			core::get_instance<core_threads_jobs_manager_t>()->queue(p);

		// wait for job to finish
		for (auto& p : list)
		{
			if (p != nullptr)
				if (!p->wait(timeout))
					return false;
		}

		// eventually throw error
		for (auto& p : list)
		{
			if (__ISNULLPTR(p))
				continue;

			// throw error
			if (p->get_status() == core_threads_jobs_base_t::status::error)
				throw_exception(job_exception, p->get_error_message());

			// job should be finished by now
			if (p->get_status() != core_threads_jobs_base_t::status::done)
				throw_exception(job_done_exception);
		}

		return true;
	}
}