/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of he MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 * 
 * This Source Code is part of the More-CPP CORE library.
 * https://www.more-cpp.com
 */
#pragma once

#include "../utils.h"

#include "../../debug/log.h"
#include "../../exception.h"

#include "../../arch/singleton.h"
#include "../../arch/library/manager.h"

#include "exception.h"

#include <stdio.h>
#include <string.h>

#include <string>
#include <unordered_map>

namespace core::utils::types
{
	/* types manager singleton class */
	class manager : public core_arch_singleton_t<manager>
	{
		friend class singleton<manager>;

	public:

		// dtor
		~manager(void)
		{
			_debug("destroying types manager");
		}

		// register C to translated type conversion
		void register_type(const std::string& ctype, const std::string& ttype)
		{
			_debug("registering type \"%s\" as \"%s\"", ctype.c_str(), ttype.c_str());

			if(this->m_forward.find(ctype) == this->m_forward.end())
				this->m_forward.emplace(std::make_pair(ctype, ttype));

			if(this->m_inverse.find(ttype) == this->m_inverse.end())
				this->m_inverse.emplace(std::make_pair(ttype, ctype));
		}

		// translate ctype
		std::string get_name_from_type(const std::string& ctype) const
		{
			auto it = this->m_forward.find(ctype);

			if (it == this->m_forward.end())
				throw_exception(cannot_resolve_typename_exception, ctype);

			return it->second;
		}

		// inverse translate
		std::string get_type_from_name(const std::string& ttype) const
		{
			auto it = this->m_inverse.find(ttype);

			if (it == this->m_inverse.end())
				throw_exception(cannot_resolve_typename_exception, ttype);

			return it->second;
		}

	private:

		// ctor
		manager(void) {}

		// lists for conversion
		std::unordered_map<std::string, std::string> m_forward, m_inverse;
	};

	/* type_auto_register class */
	class auto_register
	{
	public:
		auto_register(const std::string& ctype, const std::string& ttype)
		{
			get_instance<core::utils::types::manager>()->register_type(ctype, ttype);
		}
	};
};

// shorthands
using core_utils_types_manager_t = core::utils::types::manager;
using core_utils_types_auto_register_t = core::utils::types::auto_register;

namespace core
{
	// translate type to name
	template<typename type> std::string get_name_from_type(void)
	{
		return get_instance<core_utils_types_manager_t>()->get_name_from_type(typeid(type).name());
	}

	template<typename type> std::string get_name_from_type(const type* p)
	{
		return get_name_from_type<type>();
	}

	// decorate typename with libname
	static std::string decorate_typename(const std::string& name)
	{
		if (name.find('/') != std::string::npos)
			throw_exception(ill_formated_typename_exception, name);

		auto libname = get_instance<core_arch_library_manager_t>()->get_name();

		if (libname == "")
			return name;

		return libname + "/" + name;
	}

	// undecorate typename
	static auto split_typename(const std::string& name)
	{
		struct
		{
			std::string lib, type;
		} ret_s;

		auto tokens = tokenize(name, '/');

		if (tokens.size() == 1)
		{
			ret_s.lib = "";
			ret_s.type = tokens[0];
		}
		else if (tokens.size() == 2)
		{
			ret_s.lib = tokens[0];
			ret_s.type = tokens[1];
		}
		else
			throw_exception(ill_formated_typename_exception, name);

		return ret_s;
	}

	// return true if types matches
	static bool is_type_name(const std::string& type1, const std::string& type2, bool strict=true)
	{
		auto _type1 = split_typename(type1);
		auto _type2 = split_typename(type2);

		if (strict && _strcmpi(_type1.lib.c_str(), _type2.lib.c_str()) != 0)
			return false;

		if (_strcmpi(_type1.type.c_str(), _type2.type.c_str()) != 0)
			return false;

		return true;
	}

	template<typename type> bool is_type_name(const std::string& name)
	{
		return is_type_name(get_name_from_type<type>(), name, false);
	}

	// translate name to type
	static std::string get_type_from_name(const std::string& name)
	{
		return get_instance<core_utils_types_manager_t>()->get_type_from_name(name);
	}
};

// register typename
#define REGISTER_TYPENAME_EX(class, string, var)	core_utils_types_auto_register_t register_type_##var(typeid(class).name(), string);

#define REGISTER_TYPENAME(type)						REGISTER_TYPENAME_EX(type, #type, type);