/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of he MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 * 
 * This Source Code is part of the More-CPP CORE library.
 * https://www.more-cpp.com
 */
#include "../../../common/utils/utils.h"
#include "../../../common/exception.h"
#include "../../../common/arch/message/handler.h"
#include "../../../common/arch/message/manager.h"
#include "../../../common/arch/message/entity.h"
#include "../../../common/arch/message/base.h"
#include "../../../common/utils/test/test.h"

class core_arch_message_base : public core_arch_message_base_t {};

using goc_manager_t = core_arch_message_manager_t<core_arch_message_base>;

INITIALIZE_SINGLETON(goc_manager_t);

/* simple message class, nothing special */
class core_arch_message : public core_arch_message_base
{
public:
	core_arch_message(int value=0) : m_value(value) {}

	int get(void) const
	{
		return this->m_value;
	}

private:
	int m_value;
};

REGISTER_TYPENAME(core_arch_message);

/* message mock up entity sample class */
class core_arch_message_mockup : public core_arch_message_handler_t<core_arch_message_base>
{
public:
	core_arch_message_mockup(void)
	{
		REGISTER_MESSAGE(core_arch_message_mockup, core_arch_message);

		this->m_value = 0;
	}

	int get(void) const
	{
		return this->m_value;
	}

private:
	HANDLE_MESSAGE(core_arch_message, msg)
	{
		this->m_value = msg.get();
	}

private:
	int m_value;
};

/* message mock up entity sample class */
class core_arch_message_entity_mockup : public core_arch_message_entity_t<core_arch_message_base>
{
public:
	core_arch_message_entity_mockup(void)
	{
		REGISTER_ENTITY_MESSAGE(core_arch_message_entity_mockup, core_arch_message);

		this->m_value = 0;
	}

	int get(void) const
	{
		return this->m_value;
	}

	std::string get_other_uid(void) const
	{
		return this->m_other_uid;
	}

private:
	HANDLE_ENTITY_MESSAGE(core_arch_message, msg, sender_id)
	{
		this->m_value = msg.get();
		this->m_other_uid = sender_id;
	}

private:
	int m_value;
	std::string m_other_uid;
};

/* message tests */
class core_arch_message_test : public core_utils_test_base_t
{
public:
	virtual void run(void) const override
	{
		_comment("Message tests");

		// direct object handle
		{
			// create object
			core_arch_message_mockup obj;

			EXPECT_INT("initial value shall be 0", obj.get(), 0);

			// send message
			{
				obj.handle(core_arch_message(1));

				EXPECT_INT("value shall now be 1", obj.get(), 1);
			}
		}

		// access through message manager
		{
			auto p1 = std::make_shared<core_arch_message_entity_mockup>();
			auto p2 = std::make_shared<core_arch_message_entity_mockup>();

			EXPECT_NOTNULL("entity p1 shall not be null", p1.get());
			EXPECT_NOTNULL("entity p2 shall not be null", p2.get());

			EXPECT_INT("p1 initial value shall be 0", p1->get(), 0);
			EXPECT_INT("p2 initial value shall be 0", p2->get(), 0);

			p1->set_name("p1");
			p2->set_name("p2");

			EXPECT_STR("p1 name shall be \"p1\"", p1->name(), "p1");
			EXPECT_STR("p2 name shall be \"p2\"", p2->name(), "p2");

			EXPECT_STR("p1 name shall be \"p1\" from uid", core::get_instance<goc_manager_t>()->inv_resolve(p1->uid()), "p1");

			EXPECT_TRUE("goc manager shall have an entity \"p1\"", core::get_instance<goc_manager_t>()->has("p1"));
			EXPECT_FALSE("goc manager shall not have an entity \"p3\"", core::get_instance<goc_manager_t>()->has("p3"));

			p1->send(p2->uid(), core_arch_message(1));

			EXPECT_INT("p1 value shall still be 0", p1->get(), 0);
			EXPECT_INT("p2 value shall now be 1", p2->get(), 1);

			EXPECT_STR("p2 value's uid owner shall be p1", p2->get_other_uid(), p1->uid());
		}
	}
};

REGISTER_TEST(core_arch_message_test);