/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of he MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 * 
 * This Source Code is part of the More-CPP CORE library.
 * https://www.more-cpp.com
 */
#include <stdio.h>

#include "../../../common/utils/utils.h"
#include "../../../common/debug/log.h"
#include "../../../common/exception.h"
#include "../../../common/debug/safe.h"
#include "../../../common/utils/test/test.h"

/* templated Safe Test */
template<typename type> class core_debug_safe_test : public core_utils_test_base_t
{
public:
	virtual void run(void) const override
	{
		_comment("Safe Arithmetic Tests for %s", typeid(type).name());

		// safe_cast
		{
			if (sizeof(type) < sizeof(size_t))
				EXPECT_EXCEPTION("safe_cast for ((size_t)~0) >> 1 shall fail", [](void)
					{
						auto tmp = core::safe_type_cast<type>(((size_t)~0) >> 1);
					});

			EXPECT_NOEXCEPTION("safe_cast for (unsigned char)1 shall succeed", [](void)
				{
					auto tmp = core::safe_type_cast<type>((unsigned char)1);
				});
		}
	}
};

/* templated Safe Test */
template<typename type> class core_debug_unsigned_safe_test : public core_debug_safe_test<type>
{
public:
	virtual void run(void) const override
	{
		core_debug_safe_test<type>::run();

		// safe_unsigned_add
		{
			EXPECT_EXCEPTION("safe_unsigned_add for 1 + ~0 shall fail", [](void)
				{
					auto tmp = core::safe_unsigned_add<type>(1, ~0);
				});

			EXPECT_EXCEPTION("safe_unsigned_add for ~0 + 1 shall fail", [](void)
				{
					auto tmp = core::safe_unsigned_add<type>(~0, 1);
				});

			EXPECT_NOEXCEPTION("safe_unsigned_add for (type)1 shall succeed", [](void)
				{
					auto tmp = core::safe_unsigned_add<type>(1, 1);
				});
		}

		// safe_unsigned_sub
		{
			EXPECT_EXCEPTION("safe_unsigned_sub for 0 - 1 shall fail", [](void)
				{
					auto tmp = core::safe_unsigned_sub<type>(0, 1);
				});

			EXPECT_NOEXCEPTION("safe_unsigned_sub for 1 - 0 shall succeed", [](void)
				{
					auto tmp = core::safe_unsigned_sub<type>(1, 0);
				});
		}

		// safe_unsigned_mult
		{
			EXPECT_EXCEPTION("safe_unsigned_mult for (~0 >> 1) * 3 shall fail", [](void)
				{
					auto tmp = core::safe_unsigned_mult<type>(((type)~0) >> 1, 3);
				});

			EXPECT_EXCEPTION("safe_unsigned_mult for 3 * (~0 >> 1) shall fail", [](void)
				{
					auto tmp = core::safe_unsigned_mult<type>(3, ((type)~0) >> 1);
				});

			EXPECT_NOEXCEPTION("safe_unsigned_mult for (~0 >> 1) * 2 shall succeed", [](void)
				{
					auto tmp = core::safe_unsigned_mult<type>(((type)~0) >> 1, 2);
				});

			EXPECT_NOEXCEPTION("safe_unsigned_mult for 2 * (~0 >> 1) shall succeed", [](void)
				{
					auto tmp = core::safe_unsigned_mult<type>(2, ((type)~0) >> 1);
				});

			EXPECT_NOEXCEPTION("safe_unsigned_mult for 1 * 1 shall succeed", [](void)
				{
					auto tmp = core::safe_unsigned_mult<type>(1, 1);
				});
		}
	}
};

/* templated Safe Test */
template<typename type> class core_debug_signed_safe_test : public core_debug_safe_test<type>
{
public:
	virtual void run(void) const override
	{
		core_debug_safe_test<type>::run();
	}
};

/* templated Safe Test */
template<typename type> class core_debug_unsigned_safe_ptr_test : public core_debug_unsigned_safe_test<type>
{
public:
	virtual void run(void) const override
	{
		core_debug_unsigned_safe_test<type>::run();

		// make_address
		{
			int v;

			EXPECT_EXCEPTION("safe_address for ~0 shall fail", [&](void)
				{
					auto tmp = core::safe_address(&v, ~0);
				});

			EXPECT_NOEXCEPTION("safe_address for 0 shall succeed", [&](void)
				{
					auto tmp = core::safe_address(&v, 0);
				});
		}

		// safe_incptr
		{
			int v;

			EXPECT_EXCEPTION("safe_incptr for ~0 shall fail", [&](void)
				{
					auto tmp = core::safe_address(&v, ~0);
				});

			EXPECT_NOEXCEPTION("safe_incptr for 0 shall succeed", [&](void)
				{
					auto tmp = core::safe_address(&v, 0);
				});
		}
	}
};

REGISTER_TEST_EX(core_debug_unsigned_safe_test<size_t>, core_debug_unsigned_safe_test_unsigned_size_t);
REGISTER_TEST_EX(core_debug_signed_safe_test<ptrdiff_t>, core_debug_signed_safe_test_unsigned_ptrdiff_t);
REGISTER_TEST_EX(core_debug_unsigned_safe_ptr_test<uintptr_t>, core_debug_unsigned_safe_ptr_test_unsigned_uintptr_t);

REGISTER_TEST_EX(core_debug_signed_safe_test<wchar_t>, core_debug_signed_safe_test_wchar_t);

REGISTER_TEST_EX(core_debug_signed_safe_test<char>, core_debug_signed_safe_test_char);
REGISTER_TEST_EX(core_debug_signed_safe_test<short>, core_debug_signed_safe_test_short);
REGISTER_TEST_EX(core_debug_signed_safe_test<int>, core_debug_signed_safe_test_int);
REGISTER_TEST_EX(core_debug_signed_safe_test<long>, core_debug_signed_safe_test_long);
REGISTER_TEST_EX(core_debug_signed_safe_test<long long>, core_debug_signed_safe_test_long_long);

REGISTER_TEST_EX(core_debug_unsigned_safe_test<unsigned char>, core_debug_unsigned_safe_test_unsigned_char);
REGISTER_TEST_EX(core_debug_unsigned_safe_test<unsigned short>, core_debug_unsigned_safe_test_unsigned_short);
REGISTER_TEST_EX(core_debug_unsigned_safe_test<unsigned int>, core_debug_unsigned_safe_test_unsigned_int);
REGISTER_TEST_EX(core_debug_unsigned_safe_test<unsigned long>, core_debug_unsigned_safe_test_unsigned_long);
REGISTER_TEST_EX(core_debug_unsigned_safe_test<unsigned long long>, core_debug_unsigned_safe_test_unsigned_long_long);

REGISTER_TEST_EX(core_debug_signed_safe_test<int8_t>, core_debug_signed_safe_test_int8_t);
REGISTER_TEST_EX(core_debug_signed_safe_test<int16_t>, core_debug_signed_safe_test_int16_t);
REGISTER_TEST_EX(core_debug_signed_safe_test<int32_t>, core_debug_signed_safe_test_int32_t);
REGISTER_TEST_EX(core_debug_signed_safe_test<int64_t>, core_debug_signed_safe_test_int64_t);

#ifdef UINT128MAX
REGISTER_TEST_EX(core_debug_signed_safe_test<int128_t>, core_debug_signed_safe_test_int128_t);
#endif

REGISTER_TEST_EX(core_debug_unsigned_safe_test<uint8_t>, core_debug_unsigned_safe_test_uint8_t);
REGISTER_TEST_EX(core_debug_unsigned_safe_test<uint16_t>, core_debug_unsigned_safe_test_uint16_t);
REGISTER_TEST_EX(core_debug_unsigned_safe_test<uint32_t>, core_debug_unsigned_safe_test_uint32_t);
REGISTER_TEST_EX(core_debug_unsigned_safe_test<uint64_t>, core_debug_unsigned_safe_test_uint64_t);

#ifdef UINT128MAX
REGISTER_TEST_EX(core_debug_unsigned_safe_test<uint128_t>, core_debug_unsigned_safe_test_uint128_t);
#endif