/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of he MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 * 
 * This Source Code is part of the More-CPP CORE library.
 * https://www.more-cpp.com
 */
#include "../../../common/utils/utils.h"
#include "../../../common/debug/exdebug.h"
#include "../../../common/exception.h"
#include "../../../common/data/cryptography/aes.h"
#include "../../../common/data/encoder/hexadecimal.h"
#include "../../../common/utils/test/test.h"

/* AES test against FIPS standard */
class core_encode_aes_test : public core_utils_test_base_t
{
public:
	virtual void run(void) const override
	{
		_comment("AES encoding/decoding against FIPS standards");

		bool big_endian = false;
		auto message = core_data_encoder_hexadecimal_t::decode("00112233445566778899aabbccddeeff", big_endian);

		// AES test vectors
		auto vectors = {

			// key - expected value pairs
			std::make_pair<std::string, std::string>("000102030405060708090a0b0c0d0e0f", "69c4e0d86a7b0430d8cdb78070b4c55a"),
			std::make_pair<std::string, std::string>("000102030405060708090a0b0c0d0e0f1011121314151617", "dda97ca4864cdfe06eaf70a0ec0d7191"),
			std::make_pair<std::string, std::string>("000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f", "8ea2b7ca516745bfeafc49904b496089"),
		};

		// loop through all
		for (auto& v : vectors)
		{
			// retrieve key
			auto key = core_data_encoder_hexadecimal_t::decode(v.first, big_endian);

			char key_length[256];
			snprintf(key_length, sizeof(key_length), "%zu bits", 8 * key.size());

			// test encryption
			auto value = core_data_cryptography_aes_t::encode(key, message);

			EXPECT_ISTR(std::string(key_length) + " AES encoding", core_data_encoder_hexadecimal_t::encode(value, big_endian), v.second);

			// test decryption
			auto dec_value = core_data_cryptography_aes_t::decode(key, value);

			EXPECT_DATA(std::string(key_length) + " AES decoding", dec_value, message);
		}

		// test invalid key size
		EXPECT_EXCEPTION("invalid key size shall trigger exception during encoding", [](void)
			{
				core_data_cryptography_aes_t::encode(
					core_data_encoder_hexadecimal_t::decode("000102030405060708090a0b0c0d0e"),
					core_data_encoder_hexadecimal_t::decode("69c4e0d86a7b0430d8cdb78070b4c55a")
				);
			});

		// test invalid key size
		EXPECT_EXCEPTION("invalid key size shall trigger exception during decoding", [](void)
			{
				auto tmp = core_data_cryptography_aes_t::encode(
					core_data_encoder_hexadecimal_t::decode("000102030405060708090a0b0c0d0e0f"),
					core_data_encoder_hexadecimal_t::decode("69c4e0d86a7b0430d8cdb78070b4c55a")
				);

				core_data_cryptography_aes_t::decode(
					core_data_encoder_hexadecimal_t::decode("000102030405060708090a0b0c0d0e"),
					tmp
				);
			});

		// test invalid message size
		EXPECT_EXCEPTION("invalid message size shall trigger exception during encoding", [](void)
			{
				core_data_cryptography_aes_t::encode(
					core_data_encoder_hexadecimal_t::decode("000102030405060708090a0b0c0d0e0f"),
					core_data_encoder_hexadecimal_t::decode("69c4e0d86a7b0430d8cdb78070b4c5")
				);
			});

		// test invalid message size
		EXPECT_EXCEPTION("invalid message size shall trigger exception during decoding", [](void)
			{
				core_data_cryptography_aes_t::decode(
					core_data_encoder_hexadecimal_t::decode("000102030405060708090a0b0c0d0e0f"),
					core_data_encoder_hexadecimal_t::decode("69c4e0d86a7b0430d8cdb78070b4c5")
				);
			});
	}
};

REGISTER_TEST(core_encode_aes_test);