/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of he MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 * 
 * This Source Code is part of the More-CPP CORE library.
 * https://www.more-cpp.com
 */
#include "../../../common/utils/utils.h"
#include "../../../common/debug/exdebug.h"
#include "../../../common/exception.h"
#include "../../../common/data/hash/sha2.h"
#include "../../../common/data/encoder/hexadecimal.h"
#include "../../../common/utils/test/test.h"

/* sha256 encoding against standard test vectors */
class core_encode_sha256_test : public core_utils_test_base_t
{
public:
	virtual void run(void) const override
	{
		_comment("SHA256 hashing against NIST test vectors");

		test("", "e3b0c442 98fc1c14 9afbf4c8 996fb924 27ae41e4 649b934c a495991b 7852b855");
		test("abc", "ba7816bf 8f01cfea 414140de 5dae2223 b00361a3 96177a9c b410ff61 f20015ad");
		test("abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq", "248d6a61 d20638b8 e5c02693 0c3e6039 a33ce459 64ff2167 f6ecedd4 19db06c1");		

		// null sha256
		{
			std::string enc_expected = "e3b0c442 98fc1c14 9afbf4c8 996fb924 27ae41e4 649b934c a495991b 7852b855";
			auto encoded_string = core_data_encoder_hexadecimal_t::encode(core::sha256(), true, 4);

			EXPECT_STR(core::format("SHA256() = \"%s\"", enc_expected.c_str()), encoded_string, enc_expected);
		}

		// string version
		{
			std::string str_in = "The quick brown fox jumps over the lazy dog.";
			std::string enc_expected = "EF537F25C895BFA782526529A9B63D97AA631564D5D789C2B765448C8635FB6C";

			auto encoded_string = core_data_encoder_hexadecimal_t::encode(core::sha256(str_in));

			EXPECT_ISTR(core::format("SHA256(\"%s\") = \"%s\"", str_in.c_str(), enc_expected.c_str()), encoded_string, enc_expected);
		}

		// char* version
		{
			std::string str_in = "The quick brown fox jumps over the lazy dog.";
			std::string enc_expected = "EF537F25C895BFA782526529A9B63D97AA631564D5D789C2B765448C8635FB6C";

			auto encoded_string = core_data_encoder_hexadecimal_t::encode(core::sha256(str_in.c_str()));

			EXPECT_ISTR(core::format("SHA256(\"%s\") = \"%s\"", str_in.c_str(), enc_expected.c_str()), encoded_string, enc_expected);
		}

		// null sha256
		{
			std::string enc_expected = "e3b0c442 98fc1c14 9afbf4c8 996fb924 27ae41e4 649b934c a495991b 7852b855";
			auto encoded_string = core_data_encoder_hexadecimal_t::encode(core::sha256(nullptr), true, 4);

			EXPECT_STR(core::format("SHA256(null) = \"%s\"", enc_expected.c_str()), encoded_string, enc_expected);
		}
	}

private:
	void test(const std::string& str_in, const std::string& enc_expected) const
	{
		core_data_buffer_t in((const unsigned char*)str_in.c_str(), str_in.length());

		auto encoded_buffer = core::sha256(in);
		auto encoded_string = core_data_encoder_hexadecimal_t::encode(encoded_buffer, true, 4);

		EXPECT_STR(core::format("SHA256(\"%s\") = \"%s\"", str_in.c_str(), enc_expected.c_str()), encoded_string, enc_expected);
	}
};

REGISTER_TEST(core_encode_sha256_test);