/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of he MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 * 
 * This Source Code is part of the More-CPP CORE library.
 * https://www.more-cpp.com
 */
#include "../../../common/utils/utils.h"
#include "../../../common/debug/exdebug.h"
#include "../../../common/exception.h"
#include "../../../common/arch/notify/notify.h"
#include "../../../common/arch/notify/property.h"
#include "../../../common/utils/test/test.h"

/* notify mockup */
class core_events_notify_mockup : public core_arch_notify_impl_t
{
public:
	core_events_notify_mockup(void)
	{
		this->m_flag = false;
	}

	void bad_trigger(void)
	{
		this->notify(1);
	}

	size_t trigger(void)
	{
		return this->notify(0);
	}

	bool get(void) const
	{
		return this->m_flag;
	}

protected:
	virtual void on_notify(int event) override
	{
		if (event == 0)
			this->m_flag = true;
	}

private:
	bool m_flag;
};

/* reception object */
class core_events_notify_listenner_mockup : public core_arch_notify_impl_t
{
public:
	core_events_notify_listenner_mockup(void)
	{
		this->m_flag = false;
	}

	void register_object(core_events_notify_mockup& obj)
	{
		this->listen(obj, 0, _SELF(core_events_notify_listenner_mockup, action));
	}

	bool get(void) const
	{
		return this->m_flag;
	}

private:
	void action(void)
	{
		this->m_flag = true;
	}

private:
	bool m_flag;
};

/* self notify mockup */
class core_events_self_notify_mockup : public core_arch_notify_impl_t
{
public:
	core_events_self_notify_mockup(void)
	{
		this->m_flag = false;

		this->listen_self(0, _SELF(core_events_self_notify_mockup, action));
	}

	size_t trigger(void)
	{
		return this->notify(0);
	}

	bool get(void) const
	{
		return this->m_flag;
	}

private:
	void action(void)
	{
		this->m_flag = true;
	}

private:
	bool m_flag;
};

/* notify test */
class core_events_notify_test : public core_utils_test_base_t
{
public:
	virtual void run(void) const override
	{
		_comment("Events notify testing");

		// on_notify
		{
			core_events_notify_mockup obj;

			EXPECT_FALSE("default state shall be false", obj.get());

			obj.bad_trigger();

			EXPECT_FALSE("state after bad trigger shall still be false", obj.get());

			obj.trigger();

			EXPECT_TRUE("state after correct trigger shall be true", obj.get());
		}

		// call lambdas
		{
			core_events_notify_mockup obj;
			core_arch_notify_listenner_t m;

			int value = 0;

			m.listen(obj, 0, [&](void)
				{
					value = 1;
				});

			m.listen(obj, 1, [&](void)
				{
					value = 2;
				});

			EXPECT_INT("initial value shall be 0", value, 0);

			obj.trigger();

			EXPECT_INT("value after trigger shall be 1", value, 1);

			obj.bad_trigger();

			EXPECT_INT("value after bad trigger shall be 2", value, 2);
		}

		// invalid listenner
		{
			core_events_notify_mockup obj;
			core_arch_notify_listenner_t m;

			EXPECT_EXCEPTION("invalid listenner shall trigger exception", [&](void)
				{
					m.listen(obj, 0, nullptr);
				});
		}

		// adding multiple listenner shall also work
		{
			core_events_notify_mockup obj;
			core_arch_notify_listenner_t m;

			int value_a = 0, value_b = 0;

			m.listen(obj, 0, [&](void)
				{
					value_a = 1;
				});

			m.listen(obj, 0, [&](void)
				{
					value_b = 2;
				});

			obj.trigger();

			EXPECT_INT("value A shall be 1", value_a, 1);
			EXPECT_INT("value B shall be 2", value_b, 2);
		}

		// self notify
		{
			core_events_self_notify_mockup obj;

			EXPECT_FALSE("initial value shall be false", obj.get());

			obj.trigger();

			EXPECT_TRUE("value after trigger shall be true", obj.get());
		}

		// notify accros objects
		{
			core_events_notify_mockup source;
			core_events_notify_listenner_mockup sink;

			sink.register_object(source);

			EXPECT_FALSE("initial state shall be false", sink.get());

			source.trigger();

			EXPECT_TRUE("state after trigger shall be true", sink.get());
		}

		// unlisten
		{
			core_events_notify_mockup obj;

			// limit scope
			{
				core_arch_notify_listenner_t m;

				m.listen(obj, 0, [](void) {});

				EXPECT_INT("shall notify 1 target when in scope", obj.trigger(), (size_t)1);
			}

			EXPECT_INT("shall notify 0 target when outside scope", obj.trigger(), (size_t)0);
		}

		// object dies before listenner
		{
			core_arch_notify_listenner_t m;

			{
				core_events_notify_mockup obj;

				m.listen(obj, 0, [](void) {});
			}
		}

		// test property
		{
			core_arch_notify_property_t<bool> flag;
			core_arch_notify_listenner_t m;

			bool notified = false;

			m.listen(flag, core_arch_notify_property_t<bool>::EVENT_CHANGED, [&](void)
				{
					notified = true;
				});

			EXPECT_FALSE("try_get on empty property shall be empty", flag.try_get().has_value());

			EXPECT_EXCEPTION("calling property before it's set shall trigger exception", [&](void)
				{
					bool value = flag;
				});

			EXPECT_FALSE("notified shall initialy be false", notified);

			flag = true;

			EXPECT_TRUE("notified shall now be true", notified);
			EXPECT_TRUE("flag shall be true", flag);

			EXPECT_TRUE("try_get on valid property shall not be empty", flag.try_get().has_value());

			notified = false;

			EXPECT_FALSE("notified shall be false again", notified);

			flag = false;

			EXPECT_TRUE("notified shall now be true", notified);
			EXPECT_FALSE("flag shall now be true", flag);

			flag.clear();

			EXPECT_FALSE("try_get shall be empty after clear", flag.try_get().has_value());
		}
	}
};

REGISTER_TEST(core_events_notify_test);