/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of he MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 * 
 * This Source Code is part of the More-CPP CORE library.
 * https://www.more-cpp.com
 */
#include <stdio.h>

#include "../../../common/utils/utils.h"
#include "../../../common/exception.h"
#include "../../../common/utils/test/test.h"
#include "../../../common/resources/proxy.h"

#include "core_resources_intval.h"

/* Construct Test */
class core_resources_construct_test : public core_utils_test_base_t
{
public:
	virtual void run(void) const override
	{
		_comment("Resources Construct Test");

		// from string
		{
			const char* resname = "string://10";

			// load resource
			core_resources_proxy_t<core_resources_intval> r(resname);

			// check that it was correctly constructed
			EXPECT_FALSE("Resource shall NOT be pending", r.is_pending());
			EXPECT_NOTNULL("Resource from string shall not be null", r.get());
			EXPECT_INT("Resource from string value shall be 10", r->get(), 10);
			EXPECT_INT("0 resource under watch", core::get_instance<core_resources_manager_t>()->num_watched(), (size_t)0);
		}

		// from base64
		{
			const char* resname = "base64://MTA=";

			// load resource
			core_resources_proxy_t<core_resources_intval> r(resname);

			// check that it was correctly constructed
			EXPECT_FALSE("Resource shall NOT be pending", r.is_pending());
			EXPECT_NOTNULL("Resource from base64 shall not be null", r.get());
			EXPECT_INT("Resource from base64 value shall be 10", r->get(), 10);
			EXPECT_INT("0 resource under watch", core::get_instance<core_resources_manager_t>()->num_watched(), (size_t)0);
		}

		// from file
		{
			const char* resname = "file://test.txt";

			// write file
			{
				FILE* p = nullptr;

				fopen_s(&p, "test.txt", "wb+");

				if (p == nullptr)
					throw_exception(cannot_open_file_exception, "test.txt");

				fputs("10", p);

				fclose(p);
			}

			// load resource
			core_resources_proxy_t<core_resources_intval> r(resname);

			// check that it was correctly constructed
			EXPECT_FALSE("Resource shall NOT be pending", r.is_pending());
			EXPECT_NOTNULL("Resource from file shall not be null", r.get());
			EXPECT_INT("Resource from file value shall be 10", r->get(), 10);
			EXPECT_INT("1 resource under watch", core::get_instance<core_resources_manager_t>()->num_watched(), (size_t)1);
		}

		// invalid file shall trigger exception
		{
			const char* resname = "file://invalid";

			// load resource
			EXPECT_EXCEPTION("Invalid resource name shall trigger exception", [&](void)
				{
					core_resources_proxy_t<core_resources_intval> r(resname);
				});

			EXPECT_INT("0 resource under watch", core::get_instance<core_resources_manager_t>()->num_watched(), (size_t)0);
		}

		// default shall trigger exception
		{
			const char* resname = "invalid";

			// load resource
			EXPECT_EXCEPTION("Undecorated resource name shall trigger exception", [&](void)
				{
					core_resources_proxy_t<core_resources_intval> r(resname);
				});

			EXPECT_INT("0 resource under watch", core::get_instance<core_resources_manager_t>()->num_watched(), (size_t)0);
		}
	}
};

REGISTER_TEST(core_resources_construct_test);