/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of he MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 * 
 * This Source Code is part of the More-CPP CORE library.
 * https://www.more-cpp.com
 */
#include <stdio.h>

#include "../../../common/utils/utils.h"
#include "../../../common/exception.h"
#include "../../../common/utils/test/test.h"
#include "../../../common/threads/event/flag.h"
#include "../../../common/resources/monitor.h"

#include "core_resources_intval.h"

extern int g_num_loads;

/* watcher Test */
class core_resources_file_watcher_test : public core_utils_test_base_t
{
public:
	virtual void run(void) const override
	{
		_comment("File Watcher Test");

		const char* _file = "test.txt";

		// watch file change
		{
			// create file monitor & event
			core_threads_event_flag_t e;

			core_resources_file_change_monitor_t f([&](const std::string& filename)
				{
					e.trigger();
				});

			// initial write
			write(_file);

			// add file to watcher
			EXPECT_TRUE("File add shall be successful", f.add(_file));

			// wait a bit for timestamp
			std::this_thread::sleep_for(std::chrono::duration<double, std::ratio<1>>(0.1));

			// rewrite
			write(_file);

			// wait
			EXPECT_TRUE("File monitor shall trigger modification before timeout", e.wait(1.0));

			// try adding a second time
			EXPECT_FALSE("Second watch shall fail", f.add(_file));

			// remove
			e.reset();

			std::filesystem::remove(_file);

			// wait
			EXPECT_TRUE("File monitor shall trigger removal before timeout", e.wait(1.0));

		}

		// watch non-existing
		{
			// create file monitor & event
			core_resources_file_change_monitor_t f([](const std::string& filename) {});

			// remove file
			std::filesystem::remove(_file);

			// add file to watcher
			EXPECT_TRUE("Watch of removed file shall succeed as well", f.add(_file));
		}		
	}

private:

	// write file
	void write(const char *filename) const
	{
		FILE* p = nullptr;

		fopen_s(&p, filename, "wb+");

		if (p == nullptr)
			throw_exception(cannot_open_file_exception, "test.txt");

		fputs("Hello World!", p);

		fclose(p);
	}
};

REGISTER_TEST(core_resources_file_watcher_test);
