/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of he MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 * 
 * This Source Code is part of the More-CPP CORE library.
 * https://www.more-cpp.com
 */
#include <stdio.h>

#include "../../../common/utils/utils.h"
#include "../../../common/exception.h"
#include "../../../common/utils/test/test.h"
#include "../../../common/resources/proxy.h"
#include "../../../common/threads/event/flag.h"

#include "core_resources_intval.h"

#include <filesystem>

/* Reload Test */
class core_resources_reload_test : public core_utils_test_base_t
{
public:
	virtual void run(void) const override
	{
		_comment("Resources Reload Test");

		// auto reload
		{
			const char* resname = "file://test.txt";

			// write file
			{
				FILE* p = nullptr;

				fopen_s(&p, "test.txt", "wb+");

				if (p == nullptr)
					throw_exception(cannot_open_file_exception, "test.txt");

				fputs("10", p);

				fclose(p);
			}

			// load resource
			core_resources_proxy_t<core_resources_intval> r(resname);

			// check that it was correctly constructed
			EXPECT_NOTNULL("Resource shall not be null", r.get());
			EXPECT_INT("Resource value shall be 10", r->get(), 10);

			// rewrite file
			{
				// wait a bit for timestamp
				std::this_thread::sleep_for(std::chrono::duration<double, std::ratio<1>>(0.1));

				FILE* p = nullptr;

				fopen_s(&p, "test.txt", "wb+");

				if (p == nullptr)
					throw_exception(cannot_open_file_exception, "test.txt");

				fputs("20", p);

				fclose(p);
			}

			// wait for reload
			core_threads_event_flag_t e;

			bool status = e.do_until([&](void)
							{
								try
								{
									if (r->get() == 20)
										e.trigger();
								}
								catch (...) {}

							}, 1.0);

			// check that it was correctly constructed
			EXPECT_TRUE("Resource auto-reload shall not timeout", status);
			EXPECT_NOTNULL("Resource after auto-reload shall not be null", r.get());
			EXPECT_INT("Resource value after auto-reload shall now be 20", r->get(), 20);
		}

		// removing resource reload
		{
			const char* resname = "file://test.txt";

			// write file
			{
				FILE* p = nullptr;

				fopen_s(&p, "test.txt", "wb+");

				if (p == nullptr)
					throw_exception(cannot_open_file_exception, "test.txt");

				fputs("10", p);

				fclose(p);
			}

			// load resource
			core_resources_proxy_t<core_resources_intval> r(resname);

			// check that it was correctly constructed
			EXPECT_NOTNULL("Resource shall not be null", r.get());

			EXPECT_INT("Resource value shall be 10", r->get(), 10);

			// set null resource
			core::set_null_resource(std::make_shared<const core_resources_intval>(-1));

			// remove resource
			std::filesystem::remove("test.txt");

			// wait for reload
			core_threads_event_flag_t e;

			bool status = e.do_until([&](void)
				{
					try
					{
						if (r->get() == -1)
							e.trigger();
					}
					catch (...) {}

				}, 1.0);

			// check that it was correctly constructed
			EXPECT_TRUE("Resource auto-reload after removal shall not timeout", status);
			EXPECT_NOTNULL("Resource auto-reload after removal shall not be null", r.get());
			EXPECT_INT("Resource value after auto-reload removal shall now be -1", r->get(), -1);

			// clear null resource
			core::set_null_resource<core_resources_intval>(nullptr);
		}

		// reload invalid resource
		{
			const char* resname = "file://test.txt";

			// remove first
			std::filesystem::remove("test.txt");

			// load resource
			core_resources_async_proxy_t<core_resources_intval> r(resname);

			// check that it was correctly constructed
			EXPECT_NULL("Resource shall be null at first", r.get());

			// write file
			{
				FILE* p = nullptr;

				fopen_s(&p, "test.txt", "wb+");

				if (p == nullptr)
					throw_exception(cannot_open_file_exception, "test.txt");

				fputs("10", p);

				fclose(p);
			}

			// wait for reload
			core_threads_event_flag_t e;

			bool status = e.do_until([&](void)
				{
					try
					{
						if (r->get() == 10)
							e.trigger();
					}
					catch (...) {}

				}, 1.0);

			// check that it was correctly constructed
			EXPECT_TRUE("Resource auto-reload after initial write shall not timeout", status);
			EXPECT_NOTNULL("Resource auto-reload after initial write shall not be null", r.get());
			EXPECT_INT("Resource value after auto-reload initial write shall now be 10", r->get(), 10);
		}

		// force reload
		{
			const char* resname = "file://test.txt";

			// write file
			{
				FILE* p = nullptr;

				fopen_s(&p, "test.txt", "wb+");

				if (p == nullptr)
					throw_exception(cannot_open_file_exception, "test.txt");

				fputs("10", p);

				fclose(p);
			}

			// load resource
			core_resources_proxy_t<core_resources_intval> r(resname);

			// check that it was correctly constructed
			EXPECT_NOTNULL("Resource shall not be null", r.get());
			EXPECT_INT("Resource value shall be 10", r->get(), 10);

			// rewrite file
			{
				FILE* p = nullptr;

				fopen_s(&p, "test.txt", "wb+");

				if (p == nullptr)
					throw_exception(cannot_open_file_exception, "test.txt");

				fputs("20", p);

				fclose(p);
			}

			// reload
			r.reload();

			// check that it was correctly constructed
			EXPECT_NOTNULL("Resource after manual reload shall not be null", r.get());
			EXPECT_INT("Resource value after manual reload shall now be 20", r->get(), 20);

			// HACK: wait for resource loading termination, should not occur in real life
			// due to resource being modified just before being destroyed and then recreated
			// the reload can then occur just after the resource has been recreated but before
			// it is accessed!
			std::this_thread::sleep_for(std::chrono::duration<double, std::ratio<1>>(0.1));

			EXPECT_TRUE("Resource loading done before timeout", r.wait(1.0));
		}
	}
};

REGISTER_TEST(core_resources_reload_test);