/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of he MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 * 
 * This Source Code is part of the More-CPP CORE library.
 * https://www.more-cpp.com
 */
#include "../../../common/utils/utils.h"
#include "../../../common/exception.h"
#include "../../../common/data/buffer.h"
#include "../../../common/utils/test/test.h"

/* storage buffer tests */
class core_storage_buffer_test : public core_utils_test_base_t
{
public:
	virtual void run(void) const override
	{
		_comment("Storage buffer tests");

		// test ctors
		{
			uint32_t test = 0xfeedbeef;

			core_data_buffer_t obj((unsigned char*)&test, sizeof(test));

			auto p = (uint32_t*)obj.data();

			EXPECT_NOTNULL("pointer shall not be null", p);
			EXPECT_INT(core::format("value shall be %.8x", test), (size_t)*p, (size_t)test);

			// test ctor from copy
			{
				auto obj2 = obj;

				auto p = (uint32_t*)obj2.data();

				EXPECT_NOTNULL("pointer shall not be null", p);
				EXPECT_INT(core::format("value shall be %.8x", test), (size_t)*p, (size_t)test);
				
				EXPECT_TRUE("object copies must be equal", obj.equals(obj2));
			}

			// test ctor from move
			{
				auto obj2 = std::move(obj);

				auto p = (uint32_t*)obj2.data();

				EXPECT_NOTNULL("pointer shall not be null", p);
				EXPECT_INT(core::format("value shall be %.8x", test), (size_t)*p, (size_t)test);

				EXPECT_NULL("initial pointer shall now be null", obj.data());
				EXPECT_INT("initial data size shall now be zero", obj.size(), (size_t)0);
			}
		}

		// test allocation & clear
		{
			core_data_buffer_t obj;

			obj.allocate(8);

			EXPECT_INT("buffer size shall be 8 bytes", obj.size(), (size_t)8);
			
			// all bytes shall be zeros
			{
				bool all_zeros = true;
				auto data = obj.data();

				EXPECT_NOTNULL("pointer shall not be null", data);

				for (size_t i = 0; i < obj.size(); i++)
					if (data[i] != 0x00)
						all_zeros = false;

				EXPECT_TRUE("all bytes shall be zero", all_zeros);
			}

			// test clear
			obj.clear();

			EXPECT_NULL("pointer shall now be null", obj.data());
			EXPECT_INT("data size shall now be zero", obj.size(), (size_t)0);
		}

		// test typed allocation
		{
			core_data_buffer_t obj;

			obj.allocate<uint32_t>();

			EXPECT_INT("buffer size shall be 4 bytes", obj.size(), (size_t)4);
		}

		// test ptr
		{
			core_data_buffer_t obj;

			auto ofs1 = obj.append<uint32_t>(1);
			auto ofs2 = obj.append<uint32_t>(2);

			auto p1 = obj.ptr<uint32_t>(ofs1);
			auto p2 = obj.ptr<uint32_t>(ofs2);

			EXPECT_NOTNULL("data at offset 1 shall not be null", p1);
			EXPECT_INT("data at offset 1 shall be 1", (size_t)*p1, (size_t)1);

			EXPECT_NOTNULL("data at offset 2 shall not be null", p2);
			EXPECT_INT("data at offset 2 shall be 2", (size_t)*p2, (size_t)2);

			EXPECT_EXCEPTION("shall not be able to access data at size()", [&](void)
				{
					auto value = *obj.ptr<uint32_t>(obj.size());
				});

			EXPECT_EXCEPTION("shall not be able to access data at size() - 3 when check_bytes is set", [&](void)
				{
					auto value = *obj.ptr<uint32_t>(obj.size() - 3, 4);
				});
		}

		// test recycle
		{
			core_data_buffer_t obj;

			std::string str = "the quick brown fox jumped over the lazy dog";

			obj.append(str);
			auto size1 = obj.size();

			obj.append(str);
			auto size2 = obj.size();

			obj.append(str, true);
			auto size3 = obj.size();

			EXPECT_TRUE("size shall be affected by append when recycled is not used", size2 != size1);
			EXPECT_TRUE("size shall not be affected by append when recycled is used", size3 == size2);
		}

		// test equality
		{
			// equals
			{
				core_data_buffer_t a, b;

				a.append<uint32_t>(0xfeedbeef);
				b.append<uint32_t>(0xfeedbeef);

				EXPECT_TRUE("buffers shall be equal", a.equals(b));
			}

			// empty buffers are always equals
			{
				core_data_buffer_t a, b;

				EXPECT_TRUE("empty buffers shall be equal", a.equals(b));
			}

			// empty & non-empty buffers are always equals
			{
				core_data_buffer_t a, b;

				a.append<uint32_t>(0xfeedbeef);

				EXPECT_FALSE("empty & non-empty buffers are never equal", a.equals(b));
				EXPECT_FALSE("empty & non-empty buffers are never equal", b.equals(a));
			}
			
			// different size are never equals
			{
				core_data_buffer_t a, b;

				a.append<uint32_t>(0xfeedbeef);
				b.append<uint16_t>(0xbeef);

				EXPECT_FALSE("buffers shall not be equal", a.equals(b));
			}
		}
	}
};

REGISTER_TEST(core_storage_buffer_test);