/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of he MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 * 
 * This Source Code is part of the More-CPP CORE library.
 * https://www.more-cpp.com
 */
#include "../../../common/utils/utils.h"
#include "../../../common/exception.h"
#include "../../../common/data/buffer.h"
#include "../../../common/resources/storage/container.h"
#include "../../../common/utils/types/type.h"
#include "../../../common/utils/test/test.h"

#include "../../test-common/common.h"

#include "core_storage_mockup.h"

/* storage SPC0 tests */
class core_storage_spc0_test : public core_utils_test_base_t
{
public:
	virtual void run(void) const override
	{
		_comment("Storage SPC0 tests");

		// load/save tests
		{
			core_data_buffer_t buffer;

			// save
			{
				// create instance
				auto p = core::create_instance<core_storage_module>("core_storage_mockup");

				// set value
				p->set(20);

				EXPECT_INT("value shall be 20", p->get(), 20);

				// storage object
				core_resources_storage_container_t container;
				container.push_back("XYZ", "ABC", p);

				buffer = container.pack('SPC0');
			}

			// examinate content
			{
				core_resources_storage_container_t container;

				auto fs = container.unpack_ex(buffer);

				EXPECT_NOTNULL("filesystem reader object shall not be null", fs.get());
				EXPECT_TRUE("filesystem reader object shall have data", fs->has_data());

				auto list = fs->get_list();

				EXPECT_INT("filesystem reader object shall have exactly 1 object", list.size(), (size_t)1);

				auto object = fs->create_object();

				EXPECT_NOTNULL("storage object shall not be null", object.get());

				object->unpack(list.front());

				EXPECT_STR("owner shall be XYZ", object->get_owner_name(), "XYZ");
				EXPECT_STR("variable name shall be ABC", object->get_var_name(), "ABC");
				EXPECT_TRUE("type shall be of core_storage_mockup type", core::is_type_name<core_storage_mockup>(object->get_type_name()));

				auto children = object->get_children();

				EXPECT_INT("object shall have no children", children.size(), (size_t)0);

				auto vars = object->get_vars();

				EXPECT_INT("object shall have exactly one variable", vars.size(), (size_t)1);

				auto& v = vars.front();

				EXPECT_STR("var owner shall be of core_storage_mockup type", v->owner, core::get_name_from_type<core_storage_mockup>());
				EXPECT_STR("var name shall be \"m_data\"", v->name, "m_data");
				EXPECT_STR("var type shall be of int type", v->type, core::get_name_from_type<int>());
			}

			// reload
			{
				core_resources_storage_container_t container;

				container.unpack<core_storage_module>(buffer);

				EXPECT_INT("container shall have exactly 1 object", container.size(), (size_t)1);

				{
					std::shared_ptr<core_resources_storage_base_t> p;
					std::string owner, name;

					std::tie(owner, name, p) = container.front();

					auto obj = std::static_pointer_cast<core_storage_module>(p);

					EXPECT_NOTNULL("object pointer shall not be null", obj.get());
					EXPECT_INT("object value shall be 20 after restore", obj->get(), 20);
				}
			}
		}

		// encoding test
		{
			core_data_buffer_t buffer;

			// save
			{
				// create instance
				auto p = core::create_instance<core_storage_module>("core_storage_mockup");

				// set value
				p->set(20);

				// storage object
				core_resources_storage_container_t container;
				container.push_back(p);

				buffer = container.pack('SPC0', FLAG_ENABLE_RLE8);
			}

			// reload
			{
				core_resources_storage_container_t container;

				container.unpack<core_storage_module>(buffer);

				EXPECT_INT("encoded container shall have exactly 1 object", container.size(), (size_t)1);

				{
					std::shared_ptr<core_resources_storage_base_t> p;
					std::string owner, name;

					std::tie(owner, name, p) = container.front();

					auto obj = std::static_pointer_cast<core_storage_module>(p);

					EXPECT_NOTNULL("encoded object pointer shall not be null", obj.get());
					EXPECT_INT("encoded object value shall be 20 after restore", obj->get(), 20);
				}
			}
		}

		// security tests
		{
			core_data_buffer_t buffer1, buffer2;

			// create first buffer
			{
				auto p = core::create_instance<core_storage_module>("core_storage_mockup");

				p->set(20);

				core_resources_storage_container_t container;
				container.push_back(p);

				buffer1 = container.pack('SPC0');
			}

			// create second buffer
			{
				auto p = core::create_instance<core_storage_module>("core_storage_mockup");

				p->set(21);

				core_resources_storage_container_t container;
				container.push_back(p);

				buffer2 = container.pack('SPC0');
			}

			// the two buffer shall differ by more than 2/3rd of their size (avalanche effect)
			{
				size_t num_diff = 0;

				auto data1 = buffer1.data();
				auto data2 = buffer2.data();

				for (size_t i = 0; i < std::min(buffer1.size(), buffer2.size()); i++)
					if (data1[i] != data2[i])
						num_diff++;

				if (buffer1.size() > buffer2.size())
					num_diff += buffer1.size() - buffer2.size();
				else if (buffer2.size() > buffer1.size())
					num_diff += buffer2.size() - buffer1.size();

				EXPECT_TRUE("single bit difference shall produce avalanche effect of 8 bytes at least", num_diff >= 8);
			}

			// shall not read m_data in file
			{
				const char* lookup = "m_data";
				const size_t n = strlen(lookup);

				auto data = (char*)buffer1.data();

				bool match = false;

				for (size_t i = 0; i < (buffer1.size() - n); i++)
					if (strncmp(data + i, lookup, n) == 0)
						match = true;

				EXPECT_FALSE("shall not read \"m_data\" in buffer", match);
			}

			// changing a single bit shall not allow file reading
			{
				// single bit flip
				(*buffer1.ptr<unsigned char>(buffer1.size() >> 2)) ^= 0x01;

				// shall trigger exception
				EXPECT_EXCEPTION("changing a single bit shall trigger an exception when reading", [&](void)
					{
						core_resources_storage_container_t container;

						auto fs = container.unpack_ex(buffer1);
					});
			}
		}
	}
};

REGISTER_TEST(core_storage_spc0_test);