/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of he MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 * 
 * This Source Code is part of the More-CPP CORE library.
 * https://www.more-cpp.com
 */
#include "../../../common/utils/utils.h"
#include "../../../common/debug/exdebug.h"
#include "../../../common/exception.h"
#include "../../../common/utils/strings/fileparts.h"
#include "../../../common/utils/strings/format.h"
#include "../../../common/utils/test/test.h"

/* string test */
class core_utils_string_test : public core_utils_test_base_t
{
public:
	virtual void run(void) const override
	{
		_comment("String tests");

		// split_file_path
		{
			auto s = core::split_file_path("path1/path2\\path3/filename.csv");

			EXPECT_STR("directory shall be correctly extracted", s.directory, "path1/path2\\path3");
			EXPECT_STR("filename shall be correctly extracted", s.file, "filename.csv");
		}

		// split_file_path with empty directory
		{
			auto s = core::split_file_path("filename.csv");

			EXPECT_STR("empty directory shall be correctly extracted", s.directory, ".");
			EXPECT_STR("filename shall be correctly extracted", s.file, "filename.csv");
		}

		// split_file_path with empty filename
		{
			auto s = core::split_file_path("path/");

			EXPECT_STR("directory shall be correctly extracted", s.directory, "path");
			EXPECT_STR("empty filename shall be correctly extracted", s.file, "");
		}

		// split_file_ext
		{
			auto s = core::split_file_ext("filename.ext");

			EXPECT_STR("filename shall be correctly extracted", s.file, "filename");
			EXPECT_STR("file extension shall be correctly extracted", s.ext, "ext");
		}

		// split_file_ext with empty filename
		{
			auto s = core::split_file_ext(".ext");

			EXPECT_STR("empty filename shall be correctly extracted", s.file, "");
			EXPECT_STR("file extension shall be correctly extracted", s.ext, "ext");
		}

		// split_file_ext with empty extension
		{
			auto s = core::split_file_ext("filename");

			EXPECT_STR("filename shall be correctly extracted", s.file, "filename");
			EXPECT_STR("empty file extension shall be correctly extracted", s.ext, "");
		}

		// split_file_parts
		{
			auto s = core::split_file_parts("path/filename.ext");

			EXPECT_STR("directory shall be correctly extracted", s.directory, "path");
			EXPECT_STR("filename shall be correctly extracted", s.file, "filename");
			EXPECT_STR("file extension shall be correctly extracted", s.ext, "ext");
		}

		// build_path
		{
			auto s = core::build_path({ "path", "filename.ext" });

			EXPECT_STR("path shall be correctly constructed with slashes", s, "path/filename.ext");
		}

		// format
		EXPECT_STR("format shall display correctly", core::format("test=%d", (int)10), "test=10");

		// str_ends_with
		{
			EXPECT_TRUE("str_ends_with shall return true", core::str_ends_with("test string", "string"));
			EXPECT_FALSE("str_ends_with shall return false", core::str_ends_with("test string", "test"));
		}

		// str_begins_with
		{
			EXPECT_FALSE("str_begins_with shall return false", core::str_begins_with("test string", "string"));
			EXPECT_TRUE("str_begins_with shall return true", core::str_begins_with("test string", "test"));
		}

		// concat
		EXPECT_STR("concat shall display properly", core::concat({ "test", "string" }, " "), "test string");

		// trim
		EXPECT_STR("trim shall display properly", core::trim(" test string   "), "test string");

		// tokenize
		{
			auto tokens = core::tokenize("some \"good commands\" to   display", ' ');

			EXPECT_INT("string shall have 4 tokens", tokens.size(), (size_t)4);

			EXPECT_STR("token #1 shall be \"some\"", tokens[0], "some");
			EXPECT_STR("token #2 shall be \"good commands\"", tokens[1], "good commands");
			EXPECT_STR("token #3 shall be \"to\"", tokens[2], "to");
			EXPECT_STR("token #4 shall be \"display\"", tokens[3], "display");
		}

		// empty tokenize
		{
			auto tokens = core::tokenize("", ' ');

			EXPECT_INT("empty string shall have no token", tokens.size(), (size_t)0);
		}

		EXPECT_EXCEPTION("unterminated strings shall trigger exception", [](void)
			{
				auto tokens = core::tokenize("some \"good commands ", ' ');
			});
	}
};

REGISTER_TEST(core_utils_string_test);